# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import spl.token.instructions as spl_token
import typing

from solana.keypair import Keypair
from solana.publickey import PublicKey
from solana.rpc.types import TokenAccountOpts
from spl.token.client import Token as SplToken
from spl.token.constants import (
    ASSOCIATED_TOKEN_PROGRAM_ID,
    TOKEN_PROGRAM_ID,
)

from .accountinfo import AccountInfo
from .addressableaccount import AddressableAccount
from .combinableinstructions import CombinableInstructions
from .context import Context
from .instrumentlookup import InstrumentLookup
from .instrumentvalue import InstrumentValue
from .layouts import layouts
from .observables import Disposable
from .tokens import Instrument, Token
from .version import Version
from .wallet import Wallet
from .websocketsubscription import (
    WebSocketAccountSubscription,
    WebSocketSubscriptionManager,
)


# # 🥭 TokenAccount class
#
class TokenAccount(AddressableAccount):
    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        owner: PublicKey,
        value: InstrumentValue,
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version
        self.owner: PublicKey = owner
        self.value: InstrumentValue = value

    @staticmethod
    def derive_associated_token_address(owner: PublicKey, token: Token) -> PublicKey:
        address, _ = PublicKey.find_program_address(
            seeds=[bytes(owner), bytes(TOKEN_PROGRAM_ID), bytes(token.mint)],
            program_id=ASSOCIATED_TOKEN_PROGRAM_ID,
        )

        return address

    @staticmethod
    def create(context: Context, account: Keypair, token: Token) -> "TokenAccount":
        spl_token = SplToken(
            context.client.compatible_client, token.mint, TOKEN_PROGRAM_ID, account
        )
        owner = account.public_key
        new_account_address = spl_token.create_account(owner)
        created: typing.Optional[TokenAccount] = TokenAccount.load(
            context, new_account_address
        )
        if created is None:
            raise Exception(
                f"Newly-created SPL token account could not be found at address {new_account_address}"
            )
        return created

    @staticmethod
    def fetch_all_for_owner_and_token(
        context: Context, owner_public_key: PublicKey, token: Token
    ) -> typing.Sequence["TokenAccount"]:
        opts = TokenAccountOpts(mint=token.mint)

        token_accounts = context.client.get_token_accounts_by_owner(
            owner_public_key, opts
        )

        all_accounts: typing.List[TokenAccount] = []
        for token_account_response in token_accounts:
            account_info = AccountInfo._from_response_values(
                token_account_response["account"],
                PublicKey(token_account_response["pubkey"]),
            )
            token_account = TokenAccount.parse(account_info, token)
            all_accounts += [token_account]

        return all_accounts

    @staticmethod
    def fetch_largest_for_owner_and_token(
        context: Context, owner_public_key: PublicKey, token: Token
    ) -> typing.Optional["TokenAccount"]:
        all_accounts = TokenAccount.fetch_all_for_owner_and_token(
            context, owner_public_key, token
        )

        largest_account: typing.Optional[TokenAccount] = None
        for token_account in all_accounts:
            if (
                largest_account is None
                or token_account.value.value > largest_account.value.value
            ):
                largest_account = token_account

        return largest_account

    @staticmethod
    def find_or_create_token_address_to_use(
        context: Context, wallet: Wallet, owner: PublicKey, token: Token
    ) -> PublicKey:
        # This is a root wallet account - get the token account to use.
        associated_token_address = spl_token.get_associated_token_address(
            owner, token.mint
        )
        token_account: typing.Optional[TokenAccount] = TokenAccount.load(
            context, associated_token_address
        )
        if token_account is not None:
            # The associated token account exists so use it
            return associated_token_address

        # There is no associated token account. See if they have an old-style non-associated token account.
        largest = TokenAccount.fetch_largest_for_owner_and_token(context, owner, token)
        if largest is not None:
            # There is an old-style account so use that.
            return largest.address

        # There is no old-style token account either, so create the proper associated token account.
        signer = CombinableInstructions.from_wallet(wallet)
        create_instruction = spl_token.create_associated_token_account(
            wallet.address, owner, token.mint
        )
        create = CombinableInstructions.from_instruction(create_instruction)

        transaction_ids = (signer + create).execute(context)
        context.client.wait_for_confirmation(transaction_ids)

        return associated_token_address

    @staticmethod
    def from_layout(
        layout: typing.Any, account_info: AccountInfo, token: Token
    ) -> "TokenAccount":
        token_value = InstrumentValue(token, token.shift_to_decimals(layout.amount))
        return TokenAccount(
            account_info, Version.UNSPECIFIED, layout.owner, token_value
        )

    @staticmethod
    def parse(
        account_info: AccountInfo,
        token: typing.Optional[Token] = None,
        instrument_lookup: typing.Optional[InstrumentLookup] = None,
    ) -> "TokenAccount":
        data = account_info.data
        if len(data) != layouts.TOKEN_ACCOUNT.sizeof():
            raise Exception(
                f"Data length ({len(data)}) does not match expected size ({layouts.TOKEN_ACCOUNT.sizeof()})"
            )

        layout = layouts.TOKEN_ACCOUNT.parse(data)
        if token is None:
            if instrument_lookup is None:
                raise Exception(
                    "Neither 'Token' or 'InstrumentLookup' specified for parsing token data."
                )
            instrument: typing.Optional[Instrument] = instrument_lookup.find_by_mint(
                layout.mint
            )
            if instrument is None:
                raise Exception(
                    f"Could not find token data for token with mint '{layout.mint}'"
                )
            token = Token.ensure(instrument)

        return TokenAccount.from_layout(layout, account_info, token)

    @staticmethod
    def load(context: Context, address: PublicKey) -> typing.Optional["TokenAccount"]:
        account_info = AccountInfo.load(context, address)
        if account_info is None or (
            len(account_info.data) != layouts.TOKEN_ACCOUNT.sizeof()
        ):
            return None
        return TokenAccount.parse(
            account_info, instrument_lookup=context.instrument_lookup
        )

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["TokenAccount"], None],
    ) -> Disposable:
        token = Token.ensure(self.value.token)

        def __parser(account_info: AccountInfo) -> TokenAccount:
            return TokenAccount.parse(account_info, token=token)

        subscription = WebSocketAccountSubscription(context, self.address, __parser)
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def __str__(self) -> str:
        return (
            f"« TokenAccount {self.address}, Owner: {self.owner}, Value: {self.value} »"
        )



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import asyncio
import logging
import traceback
import typing

from decimal import Decimal
from solana.blockhash import Blockhash
from solana.rpc.commitment import Commitment, Finalized
from solana.keypair import Keypair
from solana.publickey import PublicKey
from solana.transaction import Transaction, TransactionInstruction
from solana.utils import shortvec_encoding as shortvec

from .constants import SOL_DECIMAL_DIVISOR
from .context import Context
from .instructionreporter import InstructionReporter
from .transactionmonitoring import TransactionOutcome, WebSocketTransactionMonitor
from .wallet import Wallet

_MAXIMUM_TRANSACTION_LENGTH = 1280 - 40 - 8
_PUBKEY_LENGTH = 32
_SIGNATURE_LENGTH = 64


def _split_instructions_into_chunks(
    context: Context,
    signers: typing.Sequence[Keypair],
    instructions: typing.Sequence[TransactionInstruction],
) -> typing.Sequence[typing.Sequence[TransactionInstruction]]:
    vetted_chunks: typing.List[typing.List[TransactionInstruction]] = []
    current_chunk: typing.List[TransactionInstruction] = []
    for counter, instruction in enumerate(instructions):
        instruction_size_on_its_own = CombinableInstructions.transaction_size(
            signers, [instruction]
        )
        if instruction_size_on_its_own >= _MAXIMUM_TRANSACTION_LENGTH:
            report = context.client.instruction_reporter.report(instruction)
            raise Exception(
                f"Instruction exceeds maximum size - instruction {counter} has {len(instruction.keys)} keys and creates a transaction {instruction_size_on_its_own} bytes long:\n{report}"
            )

        in_progress_chunk = current_chunk + [instruction]
        transaction_size = CombinableInstructions.transaction_size(
            signers, in_progress_chunk
        )
        if transaction_size < _MAXIMUM_TRANSACTION_LENGTH:
            current_chunk = in_progress_chunk
        else:
            vetted_chunks += [current_chunk]
            current_chunk = [instruction]

    all_chunks = vetted_chunks + [current_chunk]

    total_in_chunks = sum(map(lambda chunk: len(chunk), all_chunks))
    if total_in_chunks != len(instructions):
        raise Exception(
            f"Failed to chunk instructions. Have {total_in_chunks} instuctions in chunks. Should have {len(instructions)}."
        )

    return all_chunks


# 🥭 CombinableInstructions class
#
# This class wraps up zero or more Solana instructions and signers, and allows instances to be combined
# easily into a single instance. This instance can then be executed.
#
# This allows simple uses like, for example:
# ```
# (signers + place_orders + settle + crank).execute(context)
# ```
#
class CombinableInstructions:
    # A toggle to run both checks to ensure our calculations are accurate.
    __check_transaction_size_with_pyserum = False

    def __init__(
        self,
        signers: typing.Sequence[Keypair],
        instructions: typing.Sequence[TransactionInstruction],
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.signers: typing.Sequence[Keypair] = signers
        self.instructions: typing.Sequence[TransactionInstruction] = instructions

    @staticmethod
    def empty() -> "CombinableInstructions":
        return CombinableInstructions(signers=[], instructions=[])

    @staticmethod
    def from_signers(signers: typing.Sequence[Keypair]) -> "CombinableInstructions":
        return CombinableInstructions(signers=signers, instructions=[])

    @staticmethod
    def from_wallet(wallet: Wallet) -> "CombinableInstructions":
        return CombinableInstructions(signers=[wallet.keypair], instructions=[])

    @staticmethod
    def from_instruction(
        instruction: TransactionInstruction,
    ) -> "CombinableInstructions":
        return CombinableInstructions(signers=[], instructions=[instruction])

    # This is the expensive - but always accurate - way of calculating the size of a transaction.
    @staticmethod
    def _transaction_size_from_pyserum(
        signers: typing.Sequence[Keypair],
        instructions: typing.Sequence[TransactionInstruction],
    ) -> int:
        inspector = Transaction()
        inspector.recent_blockhash = Blockhash(str(PublicKey(3)))
        inspector.instructions.extend(instructions)
        inspector.sign(*signers)
        signed_data = inspector.serialize_message()

        length: int = len(signed_data)

        # Signature count length
        length += 1

        # Signatures
        length += len(inspector.signatures) * _SIGNATURE_LENGTH

        return length

    # This is the quicker way - just add up the sizes ourselves. It's not trivial though.

    @staticmethod
    def _calculate_transaction_size(
        signers: typing.Sequence[Keypair],
        instructions: typing.Sequence[TransactionInstruction],
    ) -> int:
        # Solana transactions have a deterministic size, but calculating it is a bit tricky.
        #
        # The transaction consists of:
        # * Message header
        # * count of number of instruction parcels
        # * 1 instruction parcel per instruction
        #
        # We're mostly interested in the number of distinct public keys used, rather than just the total number of public keys used
        # All distinct keys are passed in the message header, and 1-byte indexes are used in the instructions. That makes it all
        # a bit more compact.
        #
        # `shortvec` here is a compact representation of an integer that may take multiple bytes but only as many as it needs. What
        # we're interest in here is the count of the number of bytes the `shortvec` takes. It's normally 1, sometimes 2, but could be
        # any number.
        #
        # A public key is 32 bytes.
        #
        # A signature is 64 bytes.
        #
        # Message header is:
        # * 1 byte for number of signatures
        # * 1 byte for number of signed accounts
        # * 1 byte for number of unsigned accounts
        # * 32 bytes for recent blockhash
        # * (variable) shortvec length for number of distinct public keys
        # * (variable) 32 bytes * number  of distinct public keys
        #
        # This gives us a header size of:
        # 35 + (shortvec-length of distinct public keys) + (32 * number of distinct public keys)
        #
        # The count of number of instruction parcels is a shortvec length of the number of instructions
        #
        # Each instruction is:
        # * 1 byte for the program ID index
        # * (variable) shortvec length of the number of public keys
        # * (variable) 1 byte for the index of each account public key
        # * (variable) shortvec length of the data
        # * (variable) length of the data
        #
        # This gives us an instruction size of:
        # 1 + (shortvec-length of number of keys) + (number of keys) + (shortvec-length of the data) + (length of the data)
        #
        # This is signed, so the length is then:
        # * Message header size
        # * + each instruction size
        # * + shortvec-length of the number of signers
        # * + (number of signers * 64 bytes)
        #
        def shortvec_length(value: int) -> int:
            return len(shortvec.encode_length(value))

        program_ids = {
            instruction.program_id.to_base58() for instruction in instructions
        }
        meta_pubkeys = {
            meta.pubkey.to_base58()
            for instruction in instructions
            for meta in instruction.keys
        }
        distinct_publickeys = set.union(
            program_ids,
            meta_pubkeys,
            {signer.public_key.to_base58() for signer in signers},
        )
        num_distinct_publickeys = len(distinct_publickeys)

        # 35 + (shortvec-length of distinct public keys) + (32 * number of distinct public keys)
        header_size = (
            35
            + shortvec_length(num_distinct_publickeys)
            + (num_distinct_publickeys * _PUBKEY_LENGTH)
        )

        instruction_count_length = shortvec_length(len(instructions))

        instructions_size = 0
        for inst in instructions:
            # 1 + (shortvec-length of number of keys) + (number of keys) + (shortvec-length of the data) + (length of the data)
            instructions_size += (
                1
                + shortvec_length(len(inst.keys))
                + len(inst.keys)
                + shortvec_length(len(inst.data))
                + len(inst.data)
            )

        # Signatures
        signatures_size = 1 + (len(signers) * _SIGNATURE_LENGTH)

        # We can now calculate the total transaction size
        calculated_transaction_size = (
            header_size + instruction_count_length + instructions_size + signatures_size
        )
        return calculated_transaction_size

    # Calculate the exact size of a transaction. There's an upper limit of 1232 so we need to keep
    # all transactions below this size.
    @staticmethod
    def transaction_size(
        signers: typing.Sequence[Keypair],
        instructions: typing.Sequence[TransactionInstruction],
    ) -> int:
        calculated_transaction_size = (
            CombinableInstructions._calculate_transaction_size(signers, instructions)
        )
        if CombinableInstructions.__check_transaction_size_with_pyserum:
            pyserum_transaction_size = (
                CombinableInstructions._transaction_size_from_pyserum(
                    signers, instructions
                )
            )
            discrepancy = pyserum_transaction_size - calculated_transaction_size
            if discrepancy == 0:
                logging.debug(
                    f"txszcalc Calculated: {calculated_transaction_size}, Should be: {pyserum_transaction_size}, No Discrepancy!"
                )
            else:
                logging.error(
                    f"txszcalcerr Calculated: {calculated_transaction_size}, Should be: {pyserum_transaction_size}, Discrepancy: {discrepancy}"
                )
                return pyserum_transaction_size

        return calculated_transaction_size

    def __add__(
        self, new_instruction_data: "CombinableInstructions"
    ) -> "CombinableInstructions":
        all_signers = [*self.signers, *new_instruction_data.signers]
        all_instructions = [*self.instructions, *new_instruction_data.instructions]
        return CombinableInstructions(
            signers=all_signers, instructions=all_instructions
        )

    @property
    def is_empty(self) -> bool:
        return len(self.signers) == 0 and len(self.instructions) == 0

    def execute(
        self, context: Context, on_exception_continue: bool = False
    ) -> typing.Sequence[str]:
        chunks: typing.Sequence[
            typing.Sequence[TransactionInstruction]
        ] = _split_instructions_into_chunks(context, self.signers, self.instructions)

        if len(chunks) == 1 and len(chunks[0]) == 0:
            self._logger.info("No instructions to run.")
            return []

        if len(chunks) > 1:
            self._logger.info(f"Running instructions in {len(chunks)} transactions.")

        results: typing.List[str] = []
        for index, chunk in enumerate(chunks):
            transaction = Transaction()
            transaction.instructions.extend(chunk)
            try:
                response = context.client.send_transaction(transaction, *self.signers)
                results += [response]
            except Exception as exception:
                starts_at = sum(len(ch) for ch in chunks[0:index])
                if on_exception_continue:
                    self._logger.error(
                        f"""[{context.name}] Error executing chunk {index} (instructions {starts_at} to {starts_at + len(chunk)}) of CombinableInstruction.
{traceback.format_exc()}"""
                    )
                else:
                    raise exception

        return results

    def execute_and_confirm(
        self,
        context: Context,
        on_exception_continue: bool = False,
        maximum_rpc_resends: Decimal = Decimal(0),
        commitment: typing.Optional[Commitment] = None,
        transmission_timeout: float = 90,
    ) -> typing.Sequence[str]:
        chunks: typing.Sequence[
            typing.Sequence[TransactionInstruction]
        ] = _split_instructions_into_chunks(context, self.signers, self.instructions)

        if len(chunks) == 1 and len(chunks[0]) == 0:
            self._logger.info("No instructions to run.")
            return []

        if len(chunks) > 1:
            self._logger.info(f"Running instructions in {len(chunks)} transactions.")

        commitment = commitment or context.client.commitment
        results: typing.List[str] = []
        for index, chunk in enumerate(chunks):
            for send_count in range(int(maximum_rpc_resends) + 1):
                transaction = Transaction()
                transaction.instructions.extend(chunk)
                try:
                    signature = context.client.send_transaction(
                        transaction, *self.signers
                    )
                    outcome = WebSocketTransactionMonitor.wait_for_all(
                        context.client.cluster_ws_url,
                        [signature],
                        commitment=commitment,
                        timeout=transmission_timeout,
                    )
                    logging.debug(f"Transaction outcome:\n    {outcome}")
                    if outcome[0].outcome != TransactionOutcome.TIMEOUT:
                        results += [signature]
                        break
                    elif send_count == int(maximum_rpc_resends):
                        raise Exception(
                            f"Transaction for signature {signature} could not be found ({maximum_rpc_resends} resends)."
                        )
                except Exception as exception:
                    starts_at = sum(len(ch) for ch in chunks[0:index])
                    if on_exception_continue:
                        self._logger.error(
                            f"""[{context.name}] Error executing chunk {index} (instructions {starts_at} to {starts_at + len(chunk)}) of CombinableInstruction.
    {traceback.format_exc()}"""
                        )
                    else:
                        raise exception

        return results

    async def execute_async(self, context: Context) -> typing.Sequence[str]:
        async def __execute_chunk(
            chunk_index: int,
            offset_start: int,
            blockhash: Blockhash,
            instructions: typing.Sequence[TransactionInstruction],
        ) -> str:
            transaction = Transaction()
            transaction.instructions.extend(instructions)
            try:
                return context.client.send_transaction(
                    transaction, *self.signers, recent_blockhash=blockhash
                )
            except Exception as exception:
                self._logger.error(
                    f"""[{context.name}] Error executing chunk {chunk_index} (instructions {offset_start} to {offset_start + len(chunk)}) of CombinableInstruction.
{traceback.format_exc()}"""
                )
                raise exception

        chunks: typing.Sequence[
            typing.Sequence[TransactionInstruction]
        ] = _split_instructions_into_chunks(context, self.signers, self.instructions)

        if len(chunks) == 1 and len(chunks[0]) == 0:
            self._logger.info("No instructions to run.")
            return []

        if len(chunks) > 1:
            self._logger.info(f"Running instructions in {len(chunks)} transactions.")

        blockhash = context.client.get_recent_blockhash(commitment=Finalized)
        coroutines: typing.List[typing.Coroutine[None, None, str]] = []
        for index, chunk in enumerate(chunks):
            starts_at = sum(len(ch) for ch in chunks[0:index])
            coroutines += [__execute_chunk(index, starts_at, blockhash, chunk)]

        return await asyncio.gather(*coroutines)

    def cost_to_execute(self, context: Context) -> Decimal:
        # getFees() is deprecated and will be replaced by getFeeForMessage() at some point.
        # getFeeForMessage() is not fully available yet though.
        fee_response = context.client.compatible_client.get_fees()
        # fee_response should look like:
        # {
        #   "jsonrpc": "2.0",
        #   "result": {
        #     "context": {
        #       "slot": 1
        #     },
        #     "value": {
        #       "blockhash": "CSymwgTNX1j3E4qhKfJAUE41nBWEwXufoYryPbkde5RR",
        #       "feeCalculator": {
        #         "lamportsPerSignature": 5000
        #       },
        #       "lastValidSlot": 297,
        #       "lastValidBlockHeight": 296
        #     }
        #   },
        #   "id": 1
        # }
        number_of_signatures = len(self.signers)
        lamports_per_signature = fee_response["result"]["value"]["feeCalculator"][
            "lamportsPerSignature"
        ]

        fee_in_lamports = Decimal(number_of_signatures) * Decimal(
            lamports_per_signature
        )
        return fee_in_lamports / SOL_DECIMAL_DIVISOR

    def report(self, instruction_reporter: InstructionReporter) -> str:
        report: typing.List[str] = []
        for index, signer in enumerate(self.signers):
            report += [f"Signer[{index}]: {signer.public_key}"]

        for instruction in self.instructions:
            report += [instruction_reporter.report(instruction)]

        return "\n".join(report)

    def __str__(self) -> str:
        return self.report(InstructionReporter())

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import abc
import logging
import entropy
import typing

from ..modelstate import ModelState
from .reconciledorders import ReconciledOrders


# # 🥭 OrderReconciler class
#
# Base class for order reconciler that combines existing and desired orders into buckets inside a `ReconciledOrders`.
#
class OrderReconciler(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    @abc.abstractmethod
    def reconcile(
        self,
        model_state: ModelState,
        existing_orders: typing.Sequence[entropy.Order],
        desired_orders: typing.Sequence[entropy.Order],
    ) -> ReconciledOrders:
        raise NotImplementedError(
            "OrderReconciler.reconcile() is not implemented on the base type."
        )

    def __str__(self) -> str:
        return """« OrderReconciler »"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 NullOrderReconciler class
#
# Null implementation of OrderReconciler. Just maintains all existing orders.
#
class NullOrderReconciler(OrderReconciler):
    def __init__(self) -> None:
        super().__init__()

    def reconcile(
        self,
        _: ModelState,
        existing_orders: typing.Sequence[entropy.Order],
        desired_orders: typing.Sequence[entropy.Order],
    ) -> ReconciledOrders:
        outcomes: ReconciledOrders = ReconciledOrders()
        outcomes.to_keep = list(existing_orders)
        outcomes.to_ignore = list(desired_orders)
        return outcomes

    def __str__(self) -> str:
        return """« NullOrderReconciler »"""


# # 🥭 AlwaysReplaceOrderReconciler class
#
# Implementation of OrderReconciler that never keeps orders, just always cancels and replaces.
#
class AlwaysReplaceOrderReconciler(OrderReconciler):
    def __init__(self) -> None:
        super().__init__()

    def reconcile(
        self,
        _: ModelState,
        existing_orders: typing.Sequence[entropy.Order],
        desired_orders: typing.Sequence[entropy.Order],
    ) -> ReconciledOrders:
        outcomes: ReconciledOrders = ReconciledOrders()
        outcomes.to_cancel = list(existing_orders)
        outcomes.to_place = list(desired_orders)
        return outcomes

    def __str__(self) -> str:
        return """« AlwaysReplaceOrderReconciler »"""



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import logging
import typing

from decimal import Decimal
from solana.publickey import PublicKey

from .accountinfo import AccountInfo
from .addressableaccount import AddressableAccount
from .cache import Cache, PerpMarketCache, MarketCache
from .constants import SYSTEM_PROGRAM_ADDRESS
from .context import Context
from .instrumentlookup import InstrumentLookup
from .instrumentvalue import InstrumentValue
from .layouts import layouts
from .lotsizeconverter import LotSizeConverter, RaisingLotSizeConverter
from .marketlookup import MarketLookup
from .metadata import Metadata
from .observables import Disposable
from .tokens import Instrument, Token
from .tokenbank import TokenBank
from .version import Version
from .websocketsubscription import (
    WebSocketAccountSubscription,
    WebSocketSubscriptionManager,
)


# # 🥭 GroupSlotSpotMarket class
#
class GroupSlotSpotMarket:
    def __init__(
        self,
        address: PublicKey,
        maint_asset_weight: Decimal,
        init_asset_weight: Decimal,
        maint_liab_weight: Decimal,
        init_liab_weight: Decimal,
    ) -> None:
        self.address: PublicKey = address
        self.maint_asset_weight: Decimal = maint_asset_weight
        self.init_asset_weight: Decimal = init_asset_weight
        self.maint_liab_weight: Decimal = maint_liab_weight
        self.init_liab_weight: Decimal = init_liab_weight

    @staticmethod
    def from_layout(layout: typing.Any) -> "GroupSlotSpotMarket":
        spot_market: PublicKey = layout.spot_market
        maint_asset_weight: Decimal = round(layout.maint_asset_weight, 8)
        init_asset_weight: Decimal = round(layout.init_asset_weight, 8)
        maint_liab_weight: Decimal = round(layout.maint_liab_weight, 8)
        init_liab_weight: Decimal = round(layout.init_liab_weight, 8)
        return GroupSlotSpotMarket(
            spot_market,
            maint_asset_weight,
            init_asset_weight,
            maint_liab_weight,
            init_liab_weight,
        )

    @staticmethod
    def from_layout_or_none(
        layout: typing.Any,
    ) -> typing.Optional["GroupSlotSpotMarket"]:
        if (layout.spot_market is None) or (
            layout.spot_market == SYSTEM_PROGRAM_ADDRESS
        ):
            return None

        return GroupSlotSpotMarket.from_layout(layout)

    def __str__(self) -> str:
        return f"""« GroupSlotSpotMarket [{self.address}]
    Asset Weights:
        Initial: {self.init_asset_weight}
        Maintenance: {self.maint_asset_weight}
    Liability Weights:
        Initial: {self.init_liab_weight}
        Maintenance: {self.maint_liab_weight}
»"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 GroupSlotPerpMarket class
#
class GroupSlotPerpMarket:
    def __init__(
        self,
        address: PublicKey,
        maint_asset_weight: Decimal,
        init_asset_weight: Decimal,
        maint_liab_weight: Decimal,
        init_liab_weight: Decimal,
        liquidation_fee: Decimal,
        base_lot_size: Decimal,
        quote_lot_size: Decimal,
    ) -> None:
        self.address: PublicKey = address
        self.maint_asset_weight: Decimal = maint_asset_weight
        self.init_asset_weight: Decimal = init_asset_weight
        self.maint_liab_weight: Decimal = maint_liab_weight
        self.init_liab_weight: Decimal = init_liab_weight
        self.liquidation_fee: Decimal = liquidation_fee
        self.base_lot_size: Decimal = base_lot_size
        self.quote_lot_size: Decimal = quote_lot_size

    @staticmethod
    def from_layout(layout: typing.Any) -> "GroupSlotPerpMarket":
        perp_market: PublicKey = layout.perp_market
        maint_asset_weight: Decimal = round(layout.maint_asset_weight, 8)
        init_asset_weight: Decimal = round(layout.init_asset_weight, 8)
        maint_liab_weight: Decimal = round(layout.maint_liab_weight, 8)
        init_liab_weight: Decimal = round(layout.init_liab_weight, 8)
        liquidation_fee: Decimal = round(layout.liquidation_fee, 8)
        base_lot_size: Decimal = layout.base_lot_size
        quote_lot_size: Decimal = layout.quote_lot_size

        return GroupSlotPerpMarket(
            perp_market,
            maint_asset_weight,
            init_asset_weight,
            maint_liab_weight,
            init_liab_weight,
            liquidation_fee,
            base_lot_size,
            quote_lot_size,
        )

    @staticmethod
    def from_layout_or_none(
        layout: typing.Any,
    ) -> typing.Optional["GroupSlotPerpMarket"]:
        if (layout.perp_market is None) or (
            layout.perp_market == SYSTEM_PROGRAM_ADDRESS
        ):
            return None

        return GroupSlotPerpMarket.from_layout(layout)

    def __str__(self) -> str:
        return f"""« GroupSlotPerpMarket [{self.address}]
    Asset Weights:
        Initial: {self.init_asset_weight}
        Maintenance: {self.maint_asset_weight}
    Liability Weights:
        Initial: {self.init_liab_weight}
        Maintenance: {self.maint_liab_weight}
    Liquidation Fee: {self.liquidation_fee}
    Base Lot Size: {self.base_lot_size}
    Quote Lot Size: {self.quote_lot_size}
»"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 GroupSlot class
#
# `GroupSlot` gathers indexed slot items together instead of separate arrays.
#
class GroupSlot:
    def __init__(
        self,
        index: int,
        base_instrument: Instrument,
        base_token_bank: typing.Optional[TokenBank],
        quote_token_bank: TokenBank,
        spot_market_info: typing.Optional[GroupSlotSpotMarket],
        perp_market_info: typing.Optional[GroupSlotPerpMarket],
        perp_lot_size_converter: LotSizeConverter,
        oracle: PublicKey,
    ) -> None:
        self.index: int = index
        self.base_instrument: Instrument = base_instrument
        self.base_token_bank: typing.Optional[TokenBank] = base_token_bank
        self.quote_token_bank: TokenBank = quote_token_bank
        self.spot_market: typing.Optional[GroupSlotSpotMarket] = spot_market_info
        self.perp_market: typing.Optional[GroupSlotPerpMarket] = perp_market_info
        self.perp_lot_size_converter: LotSizeConverter = perp_lot_size_converter
        self.oracle: PublicKey = oracle

    @property
    def spot_market_symbol(self) -> str:
        return f"{self.base_instrument.symbol}/{self.quote_token_bank.token.symbol}"

    @property
    def perp_market_symbol(self) -> str:
        return f"{self.base_instrument.symbol}-PERP"

    def __str__(self) -> str:
        base_token_bank = f"{self.base_token_bank}".replace("\n", "\n        ")
        quote_token_bank = f"{self.quote_token_bank}".replace("\n", "\n        ")
        spot_market_info = f"{self.spot_market}".replace("\n", "\n        ")
        perp_market_info = f"{self.perp_market}".replace("\n", "\n        ")
        return f"""« GroupSlot[{self.index}] {self.base_instrument}
    Base Token Info:
        {base_token_bank}
    Quote Token Info:
        {quote_token_bank}
    Oracle: {self.oracle}
    Spot Market:
        {spot_market_info}
    Perp Market:
        {perp_market_info}
»"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 Group class
#
# `Group` defines root functionality for Entropy Markets.
#
class Group(AddressableAccount):
    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        name: str,
        meta_data: Metadata,
        shared_quote: TokenBank,
        slot_indices: typing.Sequence[bool],
        slots: typing.Sequence[GroupSlot],
        signer_nonce: Decimal,
        signer_key: PublicKey,
        admin: PublicKey,
        serum_program_address: PublicKey,
        cache: PublicKey,
        valid_interval: Decimal,
        insurance_vault: PublicKey,
        srm_vault: PublicKey,
        msrm_vault: PublicKey,
        fees_vault: PublicKey,
        max_entropy_accounts: Decimal,
        num_entropy_accounts: Decimal,
        referral_surcharge_centibps: Decimal,
        referral_share_centibps: Decimal,
        referral_mngo_required: Decimal,
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version
        self.name: str = name

        self.meta_data: Metadata = meta_data
        self.shared_quote: TokenBank = shared_quote
        self.slot_indices: typing.Sequence[bool] = slot_indices
        self.slots: typing.Sequence[GroupSlot] = slots
        self.signer_nonce: Decimal = signer_nonce
        self.signer_key: PublicKey = signer_key
        self.admin: PublicKey = admin
        self.serum_program_address: PublicKey = serum_program_address
        self.cache: PublicKey = cache
        self.valid_interval: Decimal = valid_interval
        self.insurance_vault: PublicKey = insurance_vault
        self.srm_vault: PublicKey = srm_vault
        self.msrm_vault: PublicKey = msrm_vault
        self.fees_vault: PublicKey = fees_vault
        self.max_entropy_accounts: Decimal = max_entropy_accounts
        self.num_entropy_accounts: Decimal = num_entropy_accounts
        self.referral_surcharge_centibps: Decimal = referral_surcharge_centibps
        self.referral_share_centibps: Decimal = referral_share_centibps
        self.referral_mngo_required: Decimal = referral_mngo_required

    @property
    def shared_quote_token(self) -> Token:
        return Token.ensure(self.shared_quote.token)

    @property
    def liquidity_incentive_token_bank(self) -> TokenBank:
        liquidity_incentive_token = "USDC"
        for token_bank in self.tokens:
            if token_bank.token.symbol_matches(liquidity_incentive_token):
                return token_bank

        raise Exception(
            f"Could not find token info for symbol {liquidity_incentive_token} in group {self.address}"
        )

    @property
    def liquidity_incentive_token(self) -> Token:
        return Token.ensure(self.liquidity_incentive_token_bank.token)

    @property
    def tokens(self) -> typing.Sequence[TokenBank]:
        return [*self.base_tokens, self.shared_quote]

    @property
    def tokens_by_index(self) -> typing.Sequence[typing.Optional[TokenBank]]:
        return [*self.base_tokens_by_index, self.shared_quote]

    @property
    def slots_by_index(self) -> typing.Sequence[typing.Optional[GroupSlot]]:
        mapped_items: typing.List[typing.Optional[GroupSlot]] = []
        slot_counter = 0
        for available in self.slot_indices:
            if available:
                mapped_items += [self.slots[slot_counter]]
                slot_counter += 1
            else:
                mapped_items += [None]

        return mapped_items

    @property
    def base_tokens(self) -> typing.Sequence[TokenBank]:
        return [
            slot.base_token_bank
            for slot in self.slots
            if slot.base_token_bank is not None
        ]

    @property
    def base_tokens_by_index(self) -> typing.Sequence[typing.Optional[TokenBank]]:
        return [
            slot.base_token_bank if slot is not None else None
            for slot in self.slots_by_index
        ]

    @property
    def oracles(self) -> typing.Sequence[PublicKey]:
        return [slot.oracle for slot in self.slots if slot.oracle is not None]

    @property
    def oracles_by_index(self) -> typing.Sequence[typing.Optional[PublicKey]]:
        return [
            slot.oracle if slot is not None else None for slot in self.slots_by_index
        ]

    @property
    def spot_markets(self) -> typing.Sequence[GroupSlotSpotMarket]:
        return [slot.spot_market for slot in self.slots if slot.spot_market is not None]

    @property
    def spot_markets_by_index(
        self,
    ) -> typing.Sequence[typing.Optional[GroupSlotSpotMarket]]:
        return [
            slot.spot_market if slot is not None else None
            for slot in self.slots_by_index
        ]

    @property
    def perp_markets(self) -> typing.Sequence[GroupSlotPerpMarket]:
        return [slot.perp_market for slot in self.slots if slot.perp_market is not None]

    @property
    def perp_markets_by_index(
        self,
    ) -> typing.Sequence[typing.Optional[GroupSlotPerpMarket]]:
        return [
            slot.perp_market if slot is not None else None
            for slot in self.slots_by_index
        ]

    @staticmethod
    def from_layout(
        layout: typing.Any,
        name: str,
        account_info: AccountInfo,
        version: Version,
        instrument_lookup: InstrumentLookup,
        market_lookup: MarketLookup,
    ) -> "Group":
        meta_data: Metadata = Metadata.from_layout(layout.meta_data)
        tokens: typing.List[typing.Optional[TokenBank]] = [
            TokenBank.from_layout_or_none(t, instrument_lookup) for t in layout.tokens
        ]

        # By convention, the shared quote token is always at the end.
        quote_token_bank: typing.Optional[TokenBank] = tokens[-1]
        if quote_token_bank is None:
            raise Exception("Could not find quote token info at end of group tokens.")
        slots: typing.List[GroupSlot] = []
        in_slots: typing.List[bool] = []
        for index in range(len(tokens) - 1):
            spot_market_info: typing.Optional[
                GroupSlotSpotMarket
            ] = GroupSlotSpotMarket.from_layout_or_none(layout.spot_markets[index])
            perp_market_info: typing.Optional[
                GroupSlotPerpMarket
            ] = GroupSlotPerpMarket.from_layout_or_none(layout.perp_markets[index])
            if (spot_market_info is None) and (perp_market_info is None):
                in_slots += [False]
            else:
                perp_lot_size_converter: LotSizeConverter = RaisingLotSizeConverter()
                base_token_bank: typing.Optional[TokenBank] = tokens[index]
                base_instrument: Instrument
                if base_token_bank is not None:
                    base_instrument = base_token_bank.token
                else:
                    # It's possible there's no underlying SPL token and we have a pure PERP market.
                    if perp_market_info is None:
                        raise Exception(
                            f"Cannot find base token or perp market info for index {index}"
                        )
                    perp_market = market_lookup.find_by_address(
                        perp_market_info.address
                    )
                    if perp_market is None:
                        in_slots += [False]
                        logging.warning(
                            f"Group cannot find base token or perp market for index {index} - {perp_market_info}"
                        )
                        continue

                    base_instrument = perp_market.base

                if perp_market_info is not None:
                    perp_lot_size_converter = LotSizeConverter(
                        base_instrument,
                        perp_market_info.base_lot_size,
                        quote_token_bank.token,
                        perp_market_info.quote_lot_size,
                    )

                oracle: PublicKey = layout.oracles[index]
                slot: GroupSlot = GroupSlot(
                    index,
                    base_instrument,
                    base_token_bank,
                    quote_token_bank,
                    spot_market_info,
                    perp_market_info,
                    perp_lot_size_converter,
                    oracle,
                )
                slots += [slot]
                in_slots += [True]

        signer_nonce: Decimal = layout.signer_nonce
        signer_key: PublicKey = layout.signer_key
        admin: PublicKey = layout.admin
        serum_program_address: PublicKey = layout.serum_program_address
        cache_address: PublicKey = layout.cache
        valid_interval: Decimal = layout.valid_interval
        insurance_vault: PublicKey = layout.insurance_vault
        srm_vault: PublicKey = layout.srm_vault
        msrm_vault: PublicKey = layout.msrm_vault
        fees_vault: PublicKey = layout.fees_vault
        max_entropy_accounts: Decimal = layout.max_entropy_accounts
        num_entropy_accounts: Decimal = layout.num_entropy_accounts

        referral_surcharge_centibps: Decimal = layout.referral_surcharge_centibps
        referral_share_centibps: Decimal = layout.referral_share_centibps
        referral_mngo_required: Decimal = layout.referral_mngo_required

        return Group(
            account_info,
            version,
            name,
            meta_data,
            quote_token_bank,
            in_slots,
            slots,
            signer_nonce,
            signer_key,
            admin,
            serum_program_address,
            cache_address,
            valid_interval,
            insurance_vault,
            srm_vault,
            msrm_vault,
            fees_vault,
            max_entropy_accounts,
            num_entropy_accounts,
            referral_surcharge_centibps,
            referral_share_centibps,
            referral_mngo_required,
        )

    @staticmethod
    def parse(
        account_info: AccountInfo,
        name: str,
        instrument_lookup: InstrumentLookup,
        market_lookup: MarketLookup,
    ) -> "Group":
        data = account_info.data
        if len(data) != layouts.GROUP.sizeof():
            raise Exception(
                f"Group data length ({len(data)}) does not match expected size ({layouts.GROUP.sizeof()})"
            )

        layout = layouts.GROUP.parse(data)
        return Group.from_layout(
            layout, name, account_info, Version.V3, instrument_lookup, market_lookup
        )

    @staticmethod
    def parse_with_context(context: Context, account_info: AccountInfo) -> "Group":
        name = context.lookup_group_name(account_info.address)
        return Group.parse(
            account_info, name, context.instrument_lookup, context.market_lookup
        )

    @staticmethod
    def load(context: Context, address: typing.Optional[PublicKey] = None) -> "Group":
        group_address: PublicKey = address or context.group_address
        account_info = AccountInfo.load(context, group_address)
        if account_info is None:
            raise Exception(f"Group account not found at address '{group_address}'")

        name = context.lookup_group_name(account_info.address)
        return Group.parse(
            account_info, name, context.instrument_lookup, context.market_lookup
        )

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["Group"], None],
    ) -> Disposable:
        def __parser(account_info: AccountInfo) -> Group:
            return Group.parse(
                account_info,
                self.name,
                context.instrument_lookup,
                context.market_lookup,
            )

        subscription = WebSocketAccountSubscription(context, self.address, __parser)
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def slot_by_spot_market_address(self, spot_market_address: PublicKey) -> GroupSlot:
        for slot in self.slots:
            if (
                slot.spot_market is not None
                and slot.spot_market.address == spot_market_address
            ):
                return slot

        raise Exception(
            f"Could not find spot market {spot_market_address} in group {self.address}"
        )

    def slot_by_perp_market_address(self, perp_market_address: PublicKey) -> GroupSlot:
        for slot in self.slots:
            if (
                slot.perp_market is not None
                and slot.perp_market.address == perp_market_address
            ):
                return slot

        raise Exception(
            f"Could not find perp market {perp_market_address} in group {self.address}"
        )

    def slot_by_instrument_or_none(
        self, instrument: Instrument
    ) -> typing.Optional[GroupSlot]:
        for slot in self.slots:
            if slot.base_instrument == instrument:
                return slot

        return None

    def slot_by_instrument(self, instrument: Instrument) -> GroupSlot:
        slot: typing.Optional[GroupSlot] = self.slot_by_instrument_or_none(instrument)
        if slot is not None:
            return slot

        raise Exception(f"Could not find slot for {instrument} in group {self.address}")

    def token_bank_by_instrument(self, instrument: Instrument) -> TokenBank:
        for token_bank in self.tokens:
            if token_bank.token == instrument:
                return token_bank

        raise Exception(f"Could not find token {instrument} in group {self.address}")

    def token_price_from_cache(
        self, cache: Cache, token: Instrument
    ) -> InstrumentValue:
        if token == self.shared_quote_token:
            # 1 USDC is always worth 1 USDC
            return InstrumentValue(self.shared_quote_token, Decimal(1))
        else:
            market_cache: MarketCache = self.market_cache_from_cache(cache, token)
            return market_cache.adjusted_price(token, self.shared_quote_token)

    def perp_market_cache_from_cache(
        self, cache: Cache, token: Instrument
    ) -> typing.Optional[PerpMarketCache]:
        market_cache: MarketCache = self.market_cache_from_cache(cache, token)
        return market_cache.perp_market

    def market_cache_from_cache_or_none(
        self, cache: Cache, instrument: Instrument
    ) -> typing.Optional[MarketCache]:
        slot: typing.Optional[GroupSlot] = self.slot_by_instrument_or_none(instrument)
        if slot is None:
            return None
        instrument_index: int = slot.index
        return cache.market_cache_for_index(instrument_index)

    def market_cache_from_cache(
        self, cache: Cache, instrument: Instrument
    ) -> MarketCache:
        market_cache: typing.Optional[
            MarketCache
        ] = self.market_cache_from_cache_or_none(cache, instrument)
        if market_cache is not None:
            return market_cache
        raise Exception(
            f"Could not find market cache for instrument {instrument.symbol}"
        )

    def fetch_cache(self, context: Context) -> Cache:
        return Cache.load(context, self.cache)

    def derive_referrer_record_address(self, context: Context, id: str) -> PublicKey:
        if not isinstance(id, str):
            raise Exception(f"Referrer ID '{id}' is not a string")

        id_bytes = id.encode("utf-8")
        if len(id_bytes) > 32:
            raise Exception(f"Referrer ID '{id}' is too long - maximum is 32 bytes")

        id_bytes_padded = id_bytes.ljust(32, b"\0")

        referrer_record_address_and_nonce: typing.Tuple[
            PublicKey, int
        ] = PublicKey.find_program_address(
            [bytes(self.address), b"ReferrerIdRecord", id_bytes_padded],
            context.entropy_program_address,
        )

        return referrer_record_address_and_nonce[0]

    def __str__(self) -> str:
        slot_count = len(self.slots)
        slots = "\n        ".join(
            [f"{item}".replace("\n", "\n        ") for item in self.slots]
        )
        return f"""« Group {self.version} [{self.address}]
    {self.meta_data}
    Name: {self.name}
    Signer [Nonce: {self.signer_nonce}]: {self.signer_key}
    Admin: {self.admin}
    DEX Program ID: {self.serum_program_address}
    Cache: {self.cache}
    Insurance Vault: {self.insurance_vault}
    SRM Vault: {self.srm_vault}
    MSRM Vault: {self.msrm_vault}
    Fees Vault: {self.fees_vault}
    Num Accounts: {self.num_entropy_accounts:,} out of {self.max_entropy_accounts:,}
    Valid Interval: {self.valid_interval}
    Referral:
        Surcharge: {self.referral_surcharge_centibps}
        Share: {self.referral_share_centibps}
        MNGO Required: {self.referral_mngo_required}
    Basket [{slot_count} markets]:
        {slots}
»"""



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import logging
import typing
import requests.exceptions
import time

from contextlib import contextmanager
from decimal import Decimal


# # 🥭 RetryWithPauses class
#
# This class takes a function and a list of durations to pause after a failed call.
#
# If the function succeeds, the resultant value is returned.
#
# If the function fails by raising an exception, the call pauses for the duration at the
# head of the list, then the head of the list is moved and the function is retried.
#
# It is retried up to the number of entries in the list of delays. If they all fail, the
# last failing exception is re-raised.
#
# This can be particularly helpful in cases where rate limits prevent further processing.
#
# This class is best used in a `with...` block using the `retry_context()` function below.
#
class RetryWithPauses:
    def __init__(
        self,
        name: str,
        func: typing.Callable[..., typing.Any],
        pauses: typing.Sequence[Decimal],
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.name: str = name
        self.func: typing.Callable[..., typing.Any] = func
        self.pauses: typing.Sequence[Decimal] = pauses

    def run(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        captured_exception: Exception
        for sleep_time_on_error in self.pauses:
            try:
                return self.func(*args, **kwargs)
            except requests.exceptions.HTTPError as exception:
                captured_exception = exception
                if exception.response is not None:
                    # "You will see HTTP respose codes 429 for too many requests
                    # or 413 for too much bandwidth."
                    if exception.response.status_code == 413:
                        self._logger.info(
                            f"Retriable call [{self.name}] rate limited (too much bandwidth) with error '{exception}'."
                        )
                    elif exception.response.status_code == 429:
                        self._logger.info(
                            f"Retriable call [{self.name}] rate limited (too many requests) with error '{exception}'."
                        )
                    else:
                        self._logger.info(
                            f"Retriable call [{self.name}] failed with unexpected HTTP error '{exception}'."
                        )
                else:
                    self._logger.info(
                        f"Retriable call [{self.name}] failed with unknown HTTP error '{exception}'."
                    )
            except Exception as exception:
                self._logger.info(
                    f"Retriable call failed [{self.name}] with error '{exception}'."
                )
                captured_exception = exception

            if sleep_time_on_error < 0:
                self._logger.info(
                    f"No more retries for [{self.name}] - propagating exception."
                )
                raise captured_exception

            self._logger.info(
                f"Will retry [{self.name}] call in {sleep_time_on_error} second(s)."
            )
            time.sleep(float(sleep_time_on_error))

        self._logger.info(
            f"End of retry loop for [{self.name}] - propagating exception."
        )
        raise captured_exception


# # 🥭 retry_context generator
#
# This is a bit of Python 'magic' to allow using the Retrier in a `with...` block.
#
# For example, this will call function `some_function(param1, param2)` up to `retry_count`
# times (7 in this case). It will only retry if the function throws an exception - the
# result of the first successful call is used to set the `result` variable:
# ```
# pauses = [Decimal(1), Decimal(2), Decimal(4)]
# with retry_context("Account Access", some_function, pauses) as retrier:
#     result = retrier.run(param1, param2)
# ```


@contextmanager
def retry_context(
    name: str, func: typing.Callable[..., typing.Any], pauses: typing.Sequence[Decimal]
) -> typing.Iterator[RetryWithPauses]:
    yield RetryWithPauses(name, func, pauses)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import enum
import typing

from solana.publickey import PublicKey

from .constants import EntropyConstants
from .instrumentlookup import InstrumentLookup
from .markets import Market
from .marketlookup import MarketLookup
from .perpmarket import PerpMarketStub
from .spotmarket import SpotMarketStub
from .tokens import Instrument, Token


class IdsJsonMarketType(enum.Enum):
    PERP = enum.auto()
    SPOT = enum.auto()

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 IdsJsonMarketLookup class
#
# This class allows us to look up market data from our ids.json configuration file.
#


class IdsJsonMarketLookup(MarketLookup):
    def __init__(self, cluster_name: str, instrument_lookup: InstrumentLookup) -> None:
        super().__init__()
        self.cluster_name: str = cluster_name
        self.instrument_lookup: InstrumentLookup = instrument_lookup

    @staticmethod
    def _from_dict(
        market_type: IdsJsonMarketType,
        entropy_program_address: PublicKey,
        group_address: PublicKey,
        data: typing.Dict[str, typing.Any],
        instrument_lookup: InstrumentLookup,
        quote_symbol: str,
    ) -> Market:
        base_symbol = data["baseSymbol"]

        base_instrument: typing.Optional[Instrument] = instrument_lookup.find_by_symbol(
            base_symbol
        )
        if base_instrument is None:
            raise Exception(
                f"Could not find base instrument with symbol '{base_symbol}'"
            )
        quote_instrument: typing.Optional[
            Instrument
        ] = instrument_lookup.find_by_symbol(quote_symbol)
        if quote_instrument is None:
            raise Exception(f"Could not find quote token with symbol '{quote_symbol}'")
        quote: Token = Token.ensure(quote_instrument)
        address = PublicKey(data["publicKey"])
        if market_type == IdsJsonMarketType.PERP:
            return PerpMarketStub(
                entropy_program_address, address, base_instrument, quote, group_address
            )
        else:
            base: Token = Token.ensure(base_instrument)
            return SpotMarketStub(
                entropy_program_address, address, base, quote, group_address
            )

    def find_by_symbol(self, symbol: str) -> typing.Optional[Market]:
        check_spots = True
        check_perps = True
        symbol = symbol.upper()
        if symbol.startswith("SPOT:"):
            symbol = symbol.split(":", 1)[1]
            check_perps = (
                False  # Skip perp markets because we're explicitly told it's a spot
            )
        elif symbol.startswith("PERP:"):
            symbol = symbol.split(":", 1)[1]
            check_spots = (
                False  # Skip spot markets because we're explicitly told it's a perp
            )

        for group in EntropyConstants["groups"]:
            if group["cluster"] == self.cluster_name:
                group_address: PublicKey = PublicKey(group["publicKey"])
                entropy_program_address: PublicKey = PublicKey(
                    group["entropyProgramId"]
                )
                if check_perps:
                    for market_data in group["perpMarkets"]:
                        if Market.symbols_match(market_data["name"], symbol):
                            return IdsJsonMarketLookup._from_dict(
                                IdsJsonMarketType.PERP,
                                entropy_program_address,
                                group_address,
                                market_data,
                                self.instrument_lookup,
                                group["quoteSymbol"],
                            )
                if check_spots:
                    for market_data in group["spotMarkets"]:
                        if Market.symbols_match(market_data["name"], symbol):
                            return IdsJsonMarketLookup._from_dict(
                                IdsJsonMarketType.SPOT,
                                entropy_program_address,
                                group_address,
                                market_data,
                                self.instrument_lookup,
                                group["quoteSymbol"],
                            )
        return None

    def find_by_address(self, address: PublicKey) -> typing.Optional[Market]:
        for group in EntropyConstants["groups"]:
            if group["cluster"] == self.cluster_name:
                group_address: PublicKey = PublicKey(group["publicKey"])
                entropy_program_address: PublicKey = PublicKey(
                    group["entropyProgramId"]
                )
                for market_data in group["perpMarkets"]:
                    if market_data["publicKey"] == str(address):
                        return IdsJsonMarketLookup._from_dict(
                            IdsJsonMarketType.PERP,
                            entropy_program_address,
                            group_address,
                            market_data,
                            self.instrument_lookup,
                            group["quoteSymbol"],
                        )
                for market_data in group["spotMarkets"]:
                    if market_data["publicKey"] == str(address):
                        return IdsJsonMarketLookup._from_dict(
                            IdsJsonMarketType.SPOT,
                            entropy_program_address,
                            group_address,
                            market_data,
                            self.instrument_lookup,
                            group["quoteSymbol"],
                        )
        return None

    def all_markets(self) -> typing.Sequence[Market]:
        markets = []
        for group in EntropyConstants["groups"]:
            if group["cluster"] == self.cluster_name:
                group_address: PublicKey = PublicKey(group["publicKey"])
                entropy_program_address: PublicKey = PublicKey(
                    group["entropyProgramId"]
                )
                for market_data in group["perpMarkets"]:
                    market = IdsJsonMarketLookup._from_dict(
                        IdsJsonMarketType.PERP,
                        entropy_program_address,
                        group_address,
                        market_data,
                        self.instrument_lookup,
                        group["quoteSymbol"],
                    )
                    markets = [market]
                for market_data in group["spotMarkets"]:
                    market = IdsJsonMarketLookup._from_dict(
                        IdsJsonMarketType.SPOT,
                        entropy_program_address,
                        group_address,
                        market_data,
                        self.instrument_lookup,
                        group["quoteSymbol"],
                    )
                    markets = [market]

        return markets



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal

from .element import Element
from ...modelstate import ModelState


# # 🥭 RoundToLotSizeElement class
#
# May modifiy an `Order`s price or quantity to ensure it's exactly aligned to the market's lot sizes.
#
class RoundToLotSizeElement(Element):
    def __init__(self) -> None:
        super().__init__()

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        pass

    @staticmethod
    def from_command_line_parameters(
        args: argparse.Namespace,
    ) -> "RoundToLotSizeElement":
        return RoundToLotSizeElement()

    def process(
        self,
        context: entropy.Context,
        model_state: ModelState,
        orders: typing.Sequence[entropy.Order],
    ) -> typing.Sequence[entropy.Order]:
        new_orders: typing.List[entropy.Order] = []
        for order in orders:
            new_price: Decimal = model_state.market.lot_size_converter.round_quote(
                order.price
            )
            new_quantity: Decimal = model_state.market.lot_size_converter.round_base(
                order.quantity
            )
            new_order: entropy.Order = order.with_update(price=new_price).with_update(
                quantity=new_quantity
            )
            if new_order.price == 0 or new_order.quantity == 0:
                self._logger.debug(
                    f"""Order removed - price or quantity rounded to zero:
    Old: {order}
    New: {new_order}"""
                )
            elif (order.price != new_order.price) or (
                order.quantity != new_order.quantity
            ):
                new_orders += [new_order]
                self._logger.debug(
                    f"""Order change - price and quantity now aligned to lot size:
    Old: {order}
    New: {new_order}"""
                )
            else:
                new_orders += [order]

        return new_orders

    def __str__(self) -> str:
        return "« RoundToLotSizeElement »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import typing

from decimal import Decimal
from pyserum.open_orders_account import OpenOrdersAccount as PySerumOpenOrdersAccount
from solana.publickey import PublicKey
from solana.rpc.types import MemcmpOpts

from .accountflags import AccountFlags
from .accountinfo import AccountInfo
from .addressableaccount import AddressableAccount
from .context import Context
from .encoding import encode_key
from .group import Group
from .layouts import layouts
from .observables import Disposable
from .placedorder import PlacedOrder
from .tokens import Token
from .version import Version
from .websocketsubscription import (
    WebSocketAccountSubscription,
    WebSocketSubscriptionManager,
)


# # 🥭 OpenOrders class
#
class OpenOrders(AddressableAccount):
    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        program_address: PublicKey,
        account_flags: AccountFlags,
        market: PublicKey,
        owner: PublicKey,
        base: Token,
        quote: Token,
        base_token_free: Decimal,
        base_token_total: Decimal,
        quote_token_free: Decimal,
        quote_token_total: Decimal,
        placed_orders: typing.Sequence[PlacedOrder],
        referrer_rebate_accrued: Decimal,
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version
        self.program_address: PublicKey = program_address
        self.account_flags: AccountFlags = account_flags
        self.market: PublicKey = market
        self.owner: PublicKey = owner
        self.base: Token = base
        self.quote: Token = quote
        self.base_token_free: Decimal = base_token_free
        self.base_token_total: Decimal = base_token_total
        self.quote_token_free: Decimal = quote_token_free
        self.quote_token_total: Decimal = quote_token_total
        self.placed_orders: typing.Sequence[PlacedOrder] = placed_orders
        self.referrer_rebate_accrued: Decimal = referrer_rebate_accrued

    @property
    def base_token_locked(self) -> Decimal:
        return self.base_token_total - self.base_token_free

    @property
    def quote_token_locked(self) -> Decimal:
        return self.quote_token_total - self.quote_token_free

    # Sometimes pyserum wants to take its own PySerumOpenOrdersAccount as a parameter (e.g. in settle_funds())
    def to_pyserum(self) -> PySerumOpenOrdersAccount:
        return PySerumOpenOrdersAccount.from_bytes(self.address, self.account_info.data)

    @staticmethod
    def from_layout(
        layout: typing.Any,
        account_info: AccountInfo,
        base: Token,
        quote: Token,
    ) -> "OpenOrders":
        account_flags = AccountFlags.from_layout(layout.account_flags)
        program_address = account_info.owner

        base_divisor = 10**base.decimals
        quote_divisor = 10**quote.decimals
        base_token_free: Decimal = layout.base_token_free / base_divisor
        base_token_total: Decimal = layout.base_token_total / base_divisor
        quote_token_free: Decimal = layout.quote_token_free / quote_divisor
        quote_token_total: Decimal = layout.quote_token_total / quote_divisor
        referrer_rebate_accrued: Decimal = (
            layout.referrer_rebate_accrued / quote_divisor
        )

        placed_orders: typing.Sequence[PlacedOrder] = []
        if account_flags.initialized:
            placed_orders = PlacedOrder.build_from_open_orders_data(
                layout.free_slot_bits,
                layout.is_bid_bits,
                layout.orders,
                layout.client_ids,
            )
        return OpenOrders(
            account_info,
            Version.UNSPECIFIED,
            program_address,
            account_flags,
            layout.market,
            layout.owner,
            base,
            quote,
            base_token_free,
            base_token_total,
            quote_token_free,
            quote_token_total,
            placed_orders,
            referrer_rebate_accrued,
        )

    @staticmethod
    def parse(account_info: AccountInfo, base: Token, quote: Token) -> "OpenOrders":
        data = account_info.data
        if len(data) != layouts.OPEN_ORDERS.sizeof():
            raise Exception(
                f"Data length ({len(data)}) does not match expected size ({layouts.OPEN_ORDERS.sizeof()})"
            )

        layout = layouts.OPEN_ORDERS.parse(data)
        return OpenOrders.from_layout(layout, account_info, base, quote)

    @staticmethod
    def load_raw_open_orders_account_infos(
        context: Context, group: Group
    ) -> typing.Dict[str, AccountInfo]:
        filters = [
            MemcmpOpts(
                offset=layouts.ACCOUNT_FLAGS.sizeof() + 37,
                bytes=encode_key(group.signer_key),
            )
        ]

        account_infos = AccountInfo.load_by_program(
            context,
            group.serum_program_address,
            data_size=layouts.OPEN_ORDERS.sizeof(),
            memcmp_opts=filters,
        )
        account_infos_by_address = {
            key: value
            for key, value in [
                (str(account_info.address), account_info)
                for account_info in account_infos
            ]
        }
        return account_infos_by_address

    @staticmethod
    def load(
        context: Context,
        address: PublicKey,
        base: Token,
        quote: Token,
    ) -> "OpenOrders":
        open_orders_account = AccountInfo.load(context, address)
        if open_orders_account is None:
            raise Exception(f"OpenOrders account not found at address '{address}'")
        return OpenOrders.parse(open_orders_account, base, quote)

    @staticmethod
    def load_for_market_and_owner(
        context: Context,
        market: PublicKey,
        owner: PublicKey,
        program_address: PublicKey,
        base: Token,
        quote: Token,
    ) -> typing.Sequence["OpenOrders"]:
        filters = [
            MemcmpOpts(
                offset=layouts.ACCOUNT_FLAGS.sizeof() + 5, bytes=encode_key(market)
            ),
            MemcmpOpts(
                offset=layouts.ACCOUNT_FLAGS.sizeof() + 37, bytes=encode_key(owner)
            ),
        ]

        account_infos = AccountInfo.load_by_program(
            context,
            program_address,
            data_size=layouts.OPEN_ORDERS.sizeof(),
            memcmp_opts=filters,
        )
        return list(
            map(
                lambda acc: OpenOrders.parse(acc, base, quote),
                account_infos,
            )
        )

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["OpenOrders"], None],
    ) -> Disposable:
        def __parser(account_info: AccountInfo) -> OpenOrders:
            return OpenOrders.parse(
                account_info,
                self.base,
                self.quote,
            )

        subscription = WebSocketAccountSubscription(context, self.address, __parser)
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def __str__(self) -> str:
        placed_orders = "\n        ".join(map(str, self.placed_orders)) or "None"

        return f"""« OpenOrders {self.base.symbol}/{self.quote.symbol} [{self.address}]:
    Flags: {self.account_flags}
    Program ID: {self.program_address}
    Market: {self.market}
    Owner: {self.owner}
    Base Token [{self.base.decimals} decimals]: {self.base_token_free:,.8f} of {self.base_token_total:,.8f}
    Quote Token [{self.quote.decimals} decimals]: {self.quote_token_free:,.8f} of {self.quote_token_total:,.8f}
    Referrer Rebate Accrued: {self.referrer_rebate_accrued:,.8f}
    Orders:
        {placed_orders}
»"""



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import logging
import numbers
import typing

from decimal import Decimal
from solana.publickey import PublicKey
from solana.rpc.types import TokenAccountOpts

from .context import Context
from .output import output
from .tokens import Instrument, Token


def _decimal_from_number(value: numbers.Number) -> Decimal:
    # Decimal constructor can only handle these Number types:
    # Union[Decimal, float, str, Tuple[int, Sequence[int], int]]
    if isinstance(value, Decimal):
        return value
    if isinstance(value, int):
        return Decimal(value)
    if isinstance(value, float):
        return Decimal(value)
    if isinstance(value, str):
        return Decimal(value)
    raise Exception(f"Cannot handle conversion of {value} to Decimal.")


# # 🥭 InstrumentValue class
#
# The `InstrumentValue` class is a simple way of keeping a token and value together, and
# displaying them nicely consistently.
#
class InstrumentValue:
    def __init__(self, token: Instrument, value: Decimal) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.token: Instrument = token
        self.value: Decimal = value
        if not isinstance(self.value, Decimal):
            raise Exception(f"Value is {type(self.value)}, not Decimal: {self.value}")

    def shift_to_native(self) -> "InstrumentValue":
        new_value = self.token.shift_to_native(self.value)
        return InstrumentValue(self.token, new_value)

    @staticmethod
    def fetch_total_value_or_none(
        context: Context, account_public_key: PublicKey, token: Token
    ) -> typing.Optional["InstrumentValue"]:
        opts = TokenAccountOpts(mint=token.mint)

        token_accounts = context.client.get_token_accounts_by_owner(
            account_public_key, opts
        )
        if len(token_accounts) == 0:
            return None

        total_value = Decimal(0)
        for token_account in token_accounts:
            token_balance: Decimal = context.client.get_token_account_balance(
                token_account["pubkey"]
            )
            total_value += token_balance

        return InstrumentValue(token, total_value)

    @staticmethod
    def fetch_total_value(
        context: Context, account_public_key: PublicKey, token: Token
    ) -> "InstrumentValue":
        value = InstrumentValue.fetch_total_value_or_none(
            context, account_public_key, token
        )
        if value is None:
            return InstrumentValue(token, Decimal(0))
        return value

    @staticmethod
    def report(
        values: typing.Sequence["InstrumentValue"],
        reporter: typing.Callable[[str], None] = output,
    ) -> None:
        for value in values:
            reporter(f"{value.value:>18,.8f} {value.token.name}")

    @staticmethod
    def find_by_symbol(
        values: typing.Sequence[typing.Optional["InstrumentValue"]], symbol: str
    ) -> "InstrumentValue":
        found = [
            value
            for value in values
            if value is not None
            and value.token is not None
            and value.token.symbol_matches(symbol)
        ]
        if len(found) == 0:
            raise Exception(f"Token '{symbol}' not found in token values: {values}")

        if len(found) > 1:
            raise Exception(
                f"Token '{symbol}' matched multiple tokens in values: {values}"
            )

        return found[0]

    @staticmethod
    def find_by_token(
        values: typing.Sequence[typing.Optional["InstrumentValue"]], token: Instrument
    ) -> "InstrumentValue":
        return InstrumentValue.find_by_symbol(values, token.symbol)

    @staticmethod
    def changes(
        before: typing.Sequence["InstrumentValue"],
        after: typing.Sequence["InstrumentValue"],
    ) -> typing.Sequence["InstrumentValue"]:
        changes: typing.List[InstrumentValue] = []
        for before_balance in before:
            after_balance = InstrumentValue.find_by_token(after, before_balance.token)
            result = InstrumentValue(
                before_balance.token, after_balance.value - before_balance.value
            )
            changes += [result]

        return changes

    def __add__(self, token_value_to_add: "InstrumentValue") -> "InstrumentValue":
        if self.token != token_value_to_add.token:
            raise Exception(
                f"Cannot add InstrumentValues from different tokens ({self.token} and {token_value_to_add.token})."
            )
        return InstrumentValue(self.token, self.value + token_value_to_add.value)

    def __sub__(self, token_value_to_subtract: "InstrumentValue") -> "InstrumentValue":
        if self.token != token_value_to_subtract.token:
            raise Exception(
                f"Cannot subtract InstrumentValues from different tokens ({self.token} and {token_value_to_subtract.token})."
            )
        return InstrumentValue(self.token, self.value - token_value_to_subtract.value)

    def __mul__(self, token_value_to_multiply: "InstrumentValue") -> "InstrumentValue":
        # Multiplying by another InstrumentValue is assumed to be a token value multiplied by a token price.
        # The result should be denominated in the currency of the price.
        return InstrumentValue(
            token_value_to_multiply.token, self.value * token_value_to_multiply.value
        )

    def __lt__(self, other: typing.Any) -> bool:
        if isinstance(other, numbers.Number):
            return self.value < _decimal_from_number(other)

        if not isinstance(other, InstrumentValue):
            return NotImplemented

        if self.token != other.token:
            raise Exception(
                f"Cannot compare token values when one token is {self.token.symbol} and the other is {other.token.symbol}."
            )
        return self.value < other.value

    def __gt__(self, other: typing.Any) -> bool:
        if isinstance(other, numbers.Number):
            return self.value > _decimal_from_number(other)

        if not isinstance(other, InstrumentValue):
            return NotImplemented

        if self.token != other.token:
            raise Exception(
                f"Cannot compare token values when one token is {self.token.symbol} and the other is {other.token.symbol}."
            )
        return self.value > other.value

    def __eq__(self, other: typing.Any) -> bool:
        if (
            isinstance(other, InstrumentValue)
            and self.token == other.token
            and self.value == other.value
        ):
            return True
        return False

    def __format__(self, format_spec: str) -> str:
        return format(str(self), format_spec)

    def __str__(self) -> str:
        name = "« Un-Named Instrument »"
        if self.token and self.token.name:
            name = self.token.name
        return f"« InstrumentValue: {self.value:>18,.8f} {name} »"

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import pandas
import typing

from dataclasses import dataclass
from decimal import Decimal
from solana.publickey import PublicKey
from solana.rpc.types import MemcmpOpts

from .accountinfo import AccountInfo
from .addressableaccount import AddressableAccount
from .cache import Cache, PerpMarketCache, RootBankCache, MarketCache
from .combinableinstructions import CombinableInstructions
from .constants import SYSTEM_PROGRAM_ADDRESS
from .context import Context
from .encoding import encode_key
from .group import Group, GroupSlot, GroupSlotPerpMarket
from .instructions import (
    build_entropy_deposit_instructions,
    build_entropy_settle_fees_instructions,
    build_entropy_settle_pnl_instructions,
    build_entropy_withdraw_instructions,
)
from .instrumentvalue import InstrumentValue
from .layouts import layouts
from .metadata import Metadata
from .observables import Disposable
from .openorders import OpenOrders
from .orders import Side
from .perpaccount import PerpAccount
from .perpmarketdetails import PerpMarketDetails
from .perpopenorders import PerpOpenOrders
from .placedorder import PlacedOrder
from .tokens import Instrument, Token
from .tokenaccount import TokenAccount
from .tokenbank import TokenBank
from .tokenoperations import build_create_associated_instructions_and_account
from .version import Version
from .wallet import Wallet
from .websocketsubscription import (
    WebSocketAccountSubscription,
    WebSocketSubscriptionManager,
)


# # 🥭 ReferrerMemory class
#
# `ReferrerMemory` stores the referrer's Entropy Account address.
#
class ReferrerMemory(AddressableAccount):
    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        meta_data: Metadata,
        referrer_entropy_account: PublicKey,
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version

        self.meta_data: Metadata = meta_data
        self.referrer_entropy_account: PublicKey = referrer_entropy_account

    @staticmethod
    def from_layout(
        layout: typing.Any, account_info: AccountInfo, version: Version
    ) -> "ReferrerMemory":
        meta_data: Metadata = Metadata.from_layout(layout.meta_data)
        referrer_entropy_account: PublicKey = layout.referrer_entropy_account

        return ReferrerMemory(
            account_info, version, meta_data, referrer_entropy_account
        )

    @staticmethod
    def parse(account_info: AccountInfo) -> "ReferrerMemory":
        data = account_info.data
        if len(data) != layouts.REFERRER_MEMORY.sizeof():
            raise Exception(
                f"ReferrerMemory data length ({len(data)}) does not match expected size ({layouts.REFERRER_MEMORY.sizeof()})"
            )

        layout = layouts.REFERRER_MEMORY.parse(data)
        return ReferrerMemory.from_layout(layout, account_info, Version.V1)

    @staticmethod
    def load_or_none(
        context: Context, address: PublicKey
    ) -> typing.Optional["ReferrerMemory"]:
        account_info = AccountInfo.load(context, address)
        if account_info is None:
            return None
        return ReferrerMemory.parse(account_info)

    @staticmethod
    def load(context: Context, address: PublicKey) -> "ReferrerMemory":
        referrer_memory: typing.Optional[ReferrerMemory] = ReferrerMemory.load_or_none(
            context, address
        )
        if referrer_memory is None:
            raise Exception(f"ReferrerMemory account not found at address '{address}'")
        return referrer_memory

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["ReferrerMemory"], None],
    ) -> Disposable:
        subscription = WebSocketAccountSubscription(
            context, self.address, ReferrerMemory.parse
        )
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def __str__(self) -> str:
        return f"""« ReferrerMemory [{self.version}] {self.address}
    {self.meta_data}
    Referrer: {self.referrer_entropy_account}
»"""


# # 🥭 AccountSlot class
#
# `AccountSlot` gathers slot items together instead of separate arrays.
#
class AccountSlot:
    def __init__(
        self,
        index: int,
        base_instrument: Instrument,
        base_token_bank: typing.Optional[TokenBank],
        quote_token_bank: TokenBank,
        raw_deposit: Decimal,
        deposit: InstrumentValue,
        raw_borrow: Decimal,
        borrow: InstrumentValue,
        spot_open_orders: typing.Optional[PublicKey],
        perp_account: typing.Optional[PerpAccount],
    ) -> None:
        self.index: int = index
        self.base_instrument: Instrument = base_instrument
        self.base_token_bank: typing.Optional[TokenBank] = base_token_bank
        self.quote_token_bank: TokenBank = quote_token_bank
        self.raw_deposit: Decimal = raw_deposit
        self.deposit: InstrumentValue = deposit
        self.raw_borrow: Decimal = raw_borrow
        self.borrow: InstrumentValue = borrow
        self.spot_open_orders: typing.Optional[PublicKey] = spot_open_orders
        self.perp_account: typing.Optional[PerpAccount] = perp_account

    @property
    def net_value(self) -> InstrumentValue:
        return self.deposit - self.borrow

    @property
    def raw_net_value(self) -> Decimal:
        return self.raw_deposit - self.raw_borrow

    def pnl(
        self,
        perp_market_cache: PerpMarketCache,
        price: InstrumentValue,
    ) -> Decimal:
        if self.perp_account is None:
            return Decimal(0)
        return self.perp_account.pnl(perp_market_cache, price)

    def __str__(self) -> str:
        perp_account: str = "None"
        if self.perp_account is not None:
            perp_account = f"{self.perp_account}".replace("\n", "\n        ")
        return f"""« AccountSlot [{self.index}] {self.base_instrument.symbol}
    Net Value:     {self.net_value}
        Deposited: {self.deposit} (raw value: {self.raw_deposit})
        Borrowed:  {self.borrow} (raw value {self.raw_borrow})
    Spot OpenOrders: {self.spot_open_orders or "None"}
    Perp Account:
        {perp_account}
»"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 Account class
#
# `Account` holds information about the account for a particular user/wallet for a particualr `Group`.
#
class Account(AddressableAccount):
    @staticmethod
    def __sum_neg(dataframe: pandas.DataFrame, name: str) -> Decimal:
        return typing.cast(Decimal, dataframe.loc[dataframe[name] < 0, name].sum())

    @staticmethod
    def __sum_pos(dataframe: pandas.DataFrame, name: str) -> Decimal:
        return typing.cast(Decimal, dataframe.loc[dataframe[name] > 0, name].sum())

    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        meta_data: Metadata,
        group_name: str,
        group_address: PublicKey,
        owner: PublicKey,
        info: str,
        shared_quote: AccountSlot,
        in_margin_basket: typing.Sequence[bool],
        slot_indices: typing.Sequence[bool],
        base_slots: typing.Sequence[AccountSlot],
        msrm_amount: Decimal,
        being_liquidated: bool,
        is_bankrupt: bool,
        advanced_orders: PublicKey,
        not_upgradable: bool,
        delegate: PublicKey,
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version

        self.meta_data: Metadata = meta_data
        self.group_name: str = group_name
        self.group_address: PublicKey = group_address
        self.owner: PublicKey = owner
        self.info: str = info
        self.shared_quote: AccountSlot = shared_quote
        self.in_margin_basket: typing.Sequence[bool] = in_margin_basket
        self.slot_indices: typing.Sequence[bool] = slot_indices
        self.base_slots: typing.Sequence[AccountSlot] = base_slots
        self.msrm_amount: Decimal = msrm_amount
        self.being_liquidated: bool = being_liquidated
        self.is_bankrupt: bool = is_bankrupt
        self.advanced_orders: PublicKey = advanced_orders
        self.not_upgradable: bool = not_upgradable
        self.delegate: PublicKey = delegate

    @property
    def shared_quote_token(self) -> Token:
        token_bank = self.shared_quote.base_token_bank
        if token_bank is None:
            raise Exception(f"Shared quote does not have a token: {self.shared_quote}")
        return Token.ensure(token_bank.token)

    @property
    def slots(self) -> typing.Sequence[AccountSlot]:
        return [*[slot for slot in self.base_slots], self.shared_quote]

    @property
    def base_slots_by_index(self) -> typing.Sequence[typing.Optional[AccountSlot]]:
        mapped_items: typing.List[typing.Optional[AccountSlot]] = []
        slot_counter = 0
        for available in self.slot_indices:
            if available:
                mapped_items += [self.base_slots[slot_counter]]
                slot_counter += 1
            else:
                mapped_items += [None]

        return mapped_items

    @property
    def slots_by_index(self) -> typing.Sequence[typing.Optional[AccountSlot]]:
        return [*self.base_slots_by_index, self.shared_quote]

    @property
    def deposits(self) -> typing.Sequence[InstrumentValue]:
        return [slot.deposit for slot in self.slots]

    @property
    def deposits_by_index(self) -> typing.Sequence[typing.Optional[InstrumentValue]]:
        return [
            slot.deposit if slot is not None else None for slot in self.slots_by_index
        ]

    @property
    def borrows(self) -> typing.Sequence[InstrumentValue]:
        return [slot.borrow for slot in self.slots]

    @property
    def borrows_by_index(self) -> typing.Sequence[typing.Optional[InstrumentValue]]:
        return [
            slot.borrow if slot is not None else None for slot in self.slots_by_index
        ]

    @property
    def net_values(self) -> typing.Sequence[InstrumentValue]:
        return [slot.net_value for slot in self.slots]

    @property
    def net_values_by_index(self) -> typing.Sequence[typing.Optional[InstrumentValue]]:
        return [
            slot.net_value if slot is not None else None for slot in self.slots_by_index
        ]

    @property
    def spot_open_orders(self) -> typing.Sequence[PublicKey]:
        return [
            slot.spot_open_orders
            for slot in self.base_slots
            if slot.spot_open_orders is not None
        ]

    @property
    def spot_open_orders_by_index(self) -> typing.Sequence[typing.Optional[PublicKey]]:
        return [
            slot.spot_open_orders if slot is not None else None
            for slot in self.base_slots_by_index
        ]

    @property
    def perp_accounts(self) -> typing.Sequence[PerpAccount]:
        return [
            slot.perp_account
            for slot in self.base_slots
            if slot.perp_account is not None
        ]

    @property
    def perp_accounts_by_index(self) -> typing.Sequence[typing.Optional[PerpAccount]]:
        return [
            slot.perp_account if slot is not None else None
            for slot in self.slots_by_index
        ]

    @staticmethod
    def from_layout(
        layout: typing.Any,
        account_info: AccountInfo,
        version: Version,
        group: Group,
        cache: Cache,
    ) -> "Account":
        meta_data = Metadata.from_layout(layout.meta_data)
        owner: PublicKey = layout.owner
        info: str = layout.info
        mngo_token = group.liquidity_incentive_token
        in_margin_basket: typing.Sequence[bool] = list(
            [bool(in_basket) for in_basket in layout.in_margin_basket]
        )
        active_in_basket: typing.List[bool] = []
        slots: typing.List[AccountSlot] = []
        placed_orders_all_markets: typing.List[typing.List[PlacedOrder]] = [
            [] for _ in range(len(group.slot_indices) - 1)
        ]
        for index, order_market in enumerate(layout.order_market):
            if order_market != 0xFF:
                side = Side.from_value(layout.order_side[index])
                id = layout.order_ids[index]
                client_id = layout.client_order_ids[index]
                placed_order = PlacedOrder(id, client_id, side)
                placed_orders_all_markets[int(order_market)] += [placed_order]

        quote_token_bank: TokenBank = group.shared_quote
        quote_token: Token = group.shared_quote_token

        for index in range(len(group.slots_by_index)):
            group_slot = group.slots_by_index[index]
            if group_slot is not None:
                instrument = group_slot.base_instrument
                token_bank = group_slot.base_token_bank
                raw_deposit: Decimal = Decimal(0)
                intrinsic_deposit: Decimal = Decimal(0)
                raw_borrow: Decimal = Decimal(0)
                intrinsic_borrow: Decimal = Decimal(0)
                if token_bank is not None:
                    raw_deposit = layout.deposits[index]
                    root_bank_cache: typing.Optional[
                        RootBankCache
                    ] = token_bank.root_bank_cache_from_cache(cache, index)
                    if root_bank_cache is None:
                        raise Exception(
                            f"No root bank cache found for token {token_bank} at index {index}"
                        )
                    intrinsic_deposit = root_bank_cache.deposit_index * raw_deposit
                    raw_borrow = layout.borrows[index]
                    intrinsic_borrow = root_bank_cache.borrow_index * raw_borrow

                deposit = InstrumentValue(
                    instrument, instrument.shift_to_decimals(intrinsic_deposit)
                )
                borrow = InstrumentValue(
                    instrument, instrument.shift_to_decimals(intrinsic_borrow)
                )

                perp_open_orders = PerpOpenOrders(placed_orders_all_markets[index])

                perp_account = PerpAccount.from_layout(
                    layout.perp_accounts[index],
                    instrument,
                    quote_token,
                    perp_open_orders,
                    group_slot.perp_lot_size_converter,
                    mngo_token,
                )
                spot_open_orders = layout.spot_open_orders[index]
                account_slot: AccountSlot = AccountSlot(
                    index,
                    instrument,
                    token_bank,
                    quote_token_bank,
                    raw_deposit,
                    deposit,
                    raw_borrow,
                    borrow,
                    spot_open_orders,
                    perp_account,
                )

                slots += [account_slot]
                active_in_basket += [True]
            else:
                active_in_basket += [False]

        quote_index: int = len(layout.deposits) - 1
        raw_quote_deposit: Decimal = layout.deposits[quote_index]
        quote_root_bank_cache: typing.Optional[
            RootBankCache
        ] = quote_token_bank.root_bank_cache_from_cache(cache, quote_index)
        if quote_root_bank_cache is None:
            raise Exception(
                f"No root bank cache found for quote token {quote_token_bank} at index {index}"
            )
        intrinsic_quote_deposit = (
            quote_root_bank_cache.deposit_index * raw_quote_deposit
        )
        quote_deposit = InstrumentValue(
            quote_token, quote_token.shift_to_decimals(intrinsic_quote_deposit)
        )
        raw_quote_borrow: Decimal = layout.borrows[quote_index]
        intrinsic_quote_borrow = quote_root_bank_cache.borrow_index * raw_quote_borrow
        quote_borrow = InstrumentValue(
            quote_token, quote_token.shift_to_decimals(intrinsic_quote_borrow)
        )
        quote: AccountSlot = AccountSlot(
            len(layout.deposits) - 1,
            quote_token_bank.token,
            quote_token_bank,
            quote_token_bank,
            raw_quote_deposit,
            quote_deposit,
            raw_quote_borrow,
            quote_borrow,
            None,
            None,
        )

        msrm_amount: Decimal = layout.msrm_amount
        being_liquidated: bool = bool(layout.being_liquidated)
        is_bankrupt: bool = bool(layout.is_bankrupt)
        advanced_orders: PublicKey = layout.advanced_orders
        not_upgradable: bool = bool(layout.not_upgradable)
        delegate: PublicKey = layout.delegate

        return Account(
            account_info,
            version,
            meta_data,
            group.name,
            group.address,
            owner,
            info,
            quote,
            in_margin_basket,
            active_in_basket,
            slots,
            msrm_amount,
            being_liquidated,
            is_bankrupt,
            advanced_orders,
            not_upgradable,
            delegate,
        )

    @staticmethod
    def parse(account_info: AccountInfo, group: Group, cache: Cache) -> "Account":
        data = account_info.data
        if len(data) != layouts.MANGO_ACCOUNT.sizeof():
            raise Exception(
                f"Account data length ({len(data)}) does not match expected size ({layouts.MANGO_ACCOUNT.sizeof()})"
            )

        layout = layouts.MANGO_ACCOUNT.parse(data)
        return Account.from_layout(layout, account_info, Version.V3, group, cache)

    @staticmethod
    def load(context: Context, address: PublicKey, group: Group) -> "Account":
        account_info = AccountInfo.load(context, address)
        if account_info is None:
            raise Exception(f"Account account not found at address '{address}'")
        cache: Cache = group.fetch_cache(context)
        return Account.parse(account_info, group, cache)

    @staticmethod
    def load_all(context: Context, group: Group) -> typing.Sequence["Account"]:
        # entropy_group is just after the METADATA, which is the first entry.
        group_offset = layouts.METADATA.sizeof()
        # owner is just after entropy_group in the layout, and it's a PublicKey which is 32 bytes.
        filters = [MemcmpOpts(offset=group_offset, bytes=encode_key(group.address))]

        account_infos = AccountInfo.load_by_program(
            context,
            context.entropy_program_address,
            memcmp_opts=filters,
            data_size=layouts.MANGO_ACCOUNT.sizeof(),
        )
        cache: Cache = group.fetch_cache(context)
        accounts: typing.List[Account] = []
        for account_info in account_infos:
            account = Account.parse(account_info, group, cache)
            accounts += [account]
        return accounts

    @staticmethod
    def load_all_for_owner(
        context: Context, owner: PublicKey, group: Group
    ) -> typing.Sequence["Account"]:
        # entropy_group is just after the METADATA, which is the first entry.
        group_offset = layouts.METADATA.sizeof()
        # owner is just after entropy_group in the layout, and it's a PublicKey which is 32 bytes.
        owner_offset = group_offset + 32
        filters = [
            MemcmpOpts(offset=group_offset, bytes=encode_key(group.address)),
            MemcmpOpts(offset=owner_offset, bytes=encode_key(owner)),
        ]

        account_infos = AccountInfo.load_by_program(
            context,
            context.entropy_program_address,
            memcmp_opts=filters,
            data_size=layouts.MANGO_ACCOUNT.sizeof(),
        )
        cache: Cache = group.fetch_cache(context)
        accounts: typing.List[Account] = []
        for account_info in account_infos:
            account = Account.parse(account_info, group, cache)
            accounts += [account]
        return accounts

    @staticmethod
    def load_all_for_delegate(
        context: Context, delegate: PublicKey, group: Group
    ) -> typing.Sequence["Account"]:
        # entropy_group is just after the METADATA, which is the first entry.
        group_offset = layouts.METADATA.sizeof()
        # delegate is a PublicKey which is 32 bytes that ends 5 bytes before the end of the layout
        delegate_offset = layouts.MANGO_ACCOUNT.sizeof() - 37
        filters = [
            MemcmpOpts(offset=group_offset, bytes=encode_key(group.address)),
            MemcmpOpts(offset=delegate_offset, bytes=encode_key(delegate)),
        ]

        account_infos = AccountInfo.load_by_program(
            context,
            context.entropy_program_address,
            memcmp_opts=filters,
            data_size=layouts.MANGO_ACCOUNT.sizeof(),
        )
        cache: Cache = group.fetch_cache(context)
        accounts: typing.List[Account] = []
        for account_info in account_infos:
            account = Account.parse(account_info, group, cache)
            accounts += [account]
        return accounts

    @staticmethod
    def load_for_owner_by_address(
        context: Context,
        owner: PublicKey,
        group: Group,
        account_address: typing.Optional[PublicKey],
    ) -> "Account":
        if account_address is not None:
            return Account.load(context, account_address, group)

        accounts: typing.Sequence[Account] = Account.load_all_for_owner(
            context, owner, group
        )
        if len(accounts) > 1:
            raise Exception(
                f"More than 1 Entropy account for owner '{owner}' and which to choose not specified."
            )

        return accounts[0]

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["Account"], None],
    ) -> Disposable:
        group: Group = Group.load(context, self.group_address)
        cache: Cache = group.fetch_cache(context)

        def __parser(account_info: AccountInfo) -> Account:
            return Account.parse(account_info, group, cache)

        subscription = WebSocketAccountSubscription(context, self.address, __parser)
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def deposit(
        self, context: Context, wallet: Wallet, value: InstrumentValue
    ) -> typing.Sequence[str]:
        token: Token = Token.ensure(value.token)
        token_account = TokenAccount.fetch_largest_for_owner_and_token(
            context, wallet.keypair.public_key, token
        )

        if token_account is None:
            raise Exception(
                f"Could not find token account for token {value.token} with owner {wallet.keypair}."
            )

        deposit_token_account = TokenAccount(
            token_account.account_info,
            token_account.version,
            token_account.owner,
            value,
        )

        group = Group.load(context, self.group_address)
        token_bank = group.token_bank_by_instrument(token)
        root_bank = token_bank.ensure_root_bank(context)
        node_bank = root_bank.pick_node_bank(context)

        signers: CombinableInstructions = CombinableInstructions.from_wallet(wallet)
        deposit = build_entropy_deposit_instructions(
            context, wallet, group, self, root_bank, node_bank, deposit_token_account
        )

        all_instructions = signers + deposit
        return all_instructions.execute(context)

    def withdraw(
        self,
        context: Context,
        wallet: Wallet,
        destination: PublicKey,
        value: InstrumentValue,
        allow_borrow: bool,
    ) -> typing.Sequence[str]:
        destination_info: typing.Optional[AccountInfo] = AccountInfo.load(
            context, destination
        )
        if destination_info is None:
            raise Exception(f"Could not find wallet at address {destination}.")

        if destination_info.owner != SYSTEM_PROGRAM_ADDRESS:
            # This is not a root wallet account
            raise Exception(
                f"Can't withdraw to address {destination} - not a wallet address."
            )

        token: Token = Token.ensure(value.token)
        token_account = TokenAccount.fetch_largest_for_owner_and_token(
            context, destination, token
        )

        withdrawal_token_account: TokenAccount
        create_ata = CombinableInstructions.empty()
        if token_account is None:
            (
                create_ata,
                token_account,
            ) = build_create_associated_instructions_and_account(
                context, wallet, destination, token
            )

        withdrawal_token_account = TokenAccount(
            token_account.account_info,
            token_account.version,
            token_account.owner,
            value,
        )

        group = Group.load(context, self.group_address)
        token_bank = group.token_bank_by_instrument(token)
        root_bank = token_bank.ensure_root_bank(context)
        node_bank = root_bank.pick_node_bank(context)

        signers: CombinableInstructions = CombinableInstructions.from_wallet(wallet)
        withdraw = build_entropy_withdraw_instructions(
            context,
            wallet,
            group,
            self,
            root_bank,
            node_bank,
            withdrawal_token_account,
            allow_borrow,
        )

        all_instructions = signers + create_ata + withdraw
        return all_instructions.execute(context)

    def redeem_all_perp_pnl(
        self,
        context: Context,
        wallet: Wallet,
        group: Group,
        cache: Cache,
    ) -> typing.Sequence[str]:
        all_accounts = Account.load_all(context, group)
        self._logger.debug(f"Fetched {len(all_accounts)} Entropy Accounts.")

        instructions = CombinableInstructions.from_wallet(wallet)
        for slot in self.slots:
            if (slot.perp_account is not None) and (not slot.perp_account.empty):
                self._logger.debug(
                    f"Redeeming perp account for {slot.base_instrument.symbol}"
                )
                price = group.token_price_from_cache(cache, slot.base_instrument)
                perp_market_cache = group.perp_market_cache_from_cache(
                    cache, slot.base_instrument
                )
                if perp_market_cache is None:
                    raise Exception(
                        f"Could not load perp market cache for {slot.base_instrument.symbol} despite having a perp account for it."
                    )

                instructions += self.build_redeem_pnl_instructions_for_market(
                    context, group, all_accounts, slot, perp_market_cache, price
                )
            else:
                self._logger.debug(f"No perp account for {slot.base_instrument.symbol}")

        self._logger.debug(
            f"Have {len(instructions.instructions)} instructions to send for all perp markets."
        )

        return instructions.execute(context)

    def redeem_pnl_for_perp_market(
        self,
        context: Context,
        wallet: Wallet,
        group: Group,
        slot: AccountSlot,
        perp_market_cache: PerpMarketCache,
        price: InstrumentValue,
    ) -> typing.Sequence[str]:
        group_slot = group.slots[slot.index]
        if group_slot.perp_market is None:
            return []

        if slot.perp_account is None:
            return []

        pnl = slot.perp_account.pnl(perp_market_cache, price)
        if pnl == 0:
            # No PnL to redeem
            return []

        self._logger.debug(
            f"Trying to settle {pnl:,.8f} {group_slot.quote_token_bank.token.symbol}."
        )

        instructions = CombinableInstructions.from_wallet(wallet)

        all_accounts = Account.load_all(context, group)
        self._logger.debug(f"Fetched {len(all_accounts)} Entropy Accounts.")

        instructions += self.build_redeem_pnl_instructions_for_market(
            context, group, all_accounts, slot, perp_market_cache, price
        )

        return instructions.execute(context)

    def build_redeem_pnl_instructions_for_market(
        self,
        context: Context,
        group: Group,
        all_accounts: typing.Sequence["Account"],
        slot: AccountSlot,
        perp_market_cache: PerpMarketCache,
        price: InstrumentValue,
    ) -> CombinableInstructions:
        group_slot = group.slots[slot.index]
        if group_slot.perp_market is None:
            self._logger.debug(f"No perp market for {slot.base_instrument.symbol}")
            return CombinableInstructions.empty()

        if slot.perp_account is None:
            self._logger.debug(f"No perp account for {slot.base_instrument.symbol}")
            return CombinableInstructions.empty()

        pnl = slot.perp_account.pnl(perp_market_cache, price)
        if pnl == 0:
            # No PnL to redeem
            return CombinableInstructions.empty()

        self._logger.debug(
            f"Trying to settle {pnl:,.8f} {group_slot.quote_token_bank.token.symbol} on {group_slot.perp_market_symbol}."
        )

        pnl_is_negative: bool = False
        if pnl < 0:
            pnl_is_negative = True

        quote_root_bank = group.shared_quote.ensure_root_bank(context)

        instructions = CombinableInstructions.empty()

        # If negative, build settle fees instruction
        if pnl_is_negative:
            perp_market = PerpMarketDetails.load(
                context, group_slot.perp_market.address, group
            )
            quote_node_bank = quote_root_bank.pick_node_bank(context)
            self._logger.debug("Adding SettleFees instruction.")
            instructions += build_entropy_settle_fees_instructions(
                context,
                group,
                perp_market,
                self,
                quote_root_bank,
                quote_node_bank,
            )

        all_accounts_except_self = filter(
            lambda acc: acc.address != self.address,
            all_accounts,
        )

        # Filter out all accounts with zero or same-signed PnL
        if pnl_is_negative:
            filtered = list(
                filter(
                    lambda acc: acc.slots[slot.index].pnl(perp_market_cache, price) > 0,
                    all_accounts_except_self,
                )
            )
        else:
            filtered = list(
                filter(
                    lambda acc: acc.slots[slot.index].pnl(perp_market_cache, price) < 0,
                    all_accounts_except_self,
                )
            )
        self._logger.debug(
            f"Removed irrelevent accounts - now have {len(filtered)} Entropy Accounts."
        )

        # Now we have a filtered list, sort it
        sorted_accounts = sorted(
            filtered,
            key=lambda acc: acc.slots[slot.index].pnl(perp_market_cache, price),
            reverse=pnl_is_negative,
        )

        # Add 5% to our collection so we're sure we're sending enough accounts even if
        # the price has shifted. (This doesn't affect the redemption value.)
        target_pnl = (pnl * Decimal("1.01")).copy_abs()

        # Loop over top N accounts until we have redeemed enough or we run out of accounts.
        collected = Decimal(0)
        for other in sorted_accounts:
            other_perp_account = other.slots[slot.index].perp_account
            if other_perp_account is not None:
                other_pnl = other_perp_account.pnl(perp_market_cache, price)

                # Build instruction
                instructions += build_entropy_settle_pnl_instructions(
                    context,
                    group,
                    group_slot,
                    self,
                    other,
                    quote_root_bank,
                )

                collected += other_pnl
                self._logger.debug(
                    f"Collected {collected:,.8f} out of {target_pnl:,.8f} - added {other_pnl:,.8f} from {other.address}"
                )

                # Add 1% to our collection so we're sure we're sending enough accounts even if
                # the price has shifted. (This doesn't affect the redemption value.)
                if collected.copy_abs() > target_pnl:
                    self._logger.debug(
                        f"Done. Have collected {collected:,.8f} to redeem against."
                    )
                    break

        self._logger.debug(
            f"Have {len(instructions.instructions)} instructions to send for {group_slot.perp_market_symbol}."
        )

        return instructions

    def slot_by_instrument_or_none(
        self, instrument: Instrument
    ) -> typing.Optional[AccountSlot]:
        for slot in self.slots:
            if slot.base_instrument == instrument:
                return slot

        return None

    def slot_by_instrument(self, instrument: Instrument) -> AccountSlot:
        slot: typing.Optional[AccountSlot] = self.slot_by_instrument_or_none(instrument)
        if slot is not None:
            return slot

        raise Exception(f"Could not find token {instrument} in account {self.address}")

    def slot_by_spot_open_orders_or_none(
        self, spot_open_orders: PublicKey
    ) -> typing.Optional[AccountSlot]:
        for slot in self.slots:
            if slot.spot_open_orders == spot_open_orders:
                return slot

        return None

    def slot_by_spot_open_orders(self, spot_open_orders: PublicKey) -> AccountSlot:
        slot: typing.Optional[AccountSlot] = self.slot_by_spot_open_orders_or_none(
            spot_open_orders
        )
        if slot is not None:
            return slot

        raise Exception(
            f"Could not find spot open orders {spot_open_orders} in account {self.address}"
        )

    def load_all_spot_open_orders(
        self, context: Context
    ) -> typing.Dict[str, OpenOrders]:
        spot_open_orders_account_infos = AccountInfo.load_multiple(
            context, self.spot_open_orders
        )
        spot_open_orders_account_infos_by_address = {
            str(account_info.address): account_info
            for account_info in spot_open_orders_account_infos
        }
        spot_open_orders: typing.Dict[str, OpenOrders] = {}
        for slot in self.base_slots:
            if slot.spot_open_orders is not None:
                account_info = spot_open_orders_account_infos_by_address[
                    str(slot.spot_open_orders)
                ]
                oo = OpenOrders.parse(
                    account_info,
                    Token.ensure(slot.base_instrument),
                    Token.ensure(self.shared_quote.base_instrument),
                )
                spot_open_orders[str(slot.spot_open_orders)] = oo
        return spot_open_orders

    def update_spot_open_orders_for_market(
        self, spot_market_index: int, spot_open_orders: PublicKey
    ) -> None:
        item_to_update = self.slots_by_index[spot_market_index]
        if item_to_update is None:
            raise Exception(
                f"Could not find AccountBasketItem in Account {self.address} at index {spot_market_index}."
            )
        item_to_update.spot_open_orders = spot_open_orders

    def derive_referrer_memory_address(self, context: Context) -> PublicKey:
        referrer_memory_address_and_nonce: typing.Tuple[
            PublicKey, int
        ] = PublicKey.find_program_address(
            [bytes(self.address), b"ReferrerMemory"], context.entropy_program_address
        )

        return referrer_memory_address_and_nonce[0]

    def fetch_default_referrer(
        self, context: Context
    ) -> typing.Optional[ReferrerMemory]:
        referrer_memory_address: PublicKey = self.derive_referrer_memory_address(
            context
        )
        return ReferrerMemory.load_or_none(context, referrer_memory_address)

    def to_dataframe(
        self,
        group: Group,
        all_spot_open_orders: typing.Dict[str, OpenOrders],
        cache: Cache,
    ) -> pandas.DataFrame:
        asset_data = []
        for slot in self.slots:
            market_cache: typing.Optional[
                MarketCache
            ] = group.market_cache_from_cache_or_none(cache, slot.base_instrument)
            price: InstrumentValue = group.token_price_from_cache(
                cache, slot.base_instrument
            )

            spot_open_orders: typing.Optional[OpenOrders] = None
            spot_health_base: Decimal = slot.net_value.value
            spot_health_quote: Decimal = Decimal(0)
            spot_bids_base_net: Decimal = Decimal(0)
            spot_asks_base_net: Decimal = Decimal(0)
            if slot.spot_open_orders is not None:
                spot_open_orders = all_spot_open_orders[str(slot.spot_open_orders)]
                if spot_open_orders is None:
                    raise Exception(
                        f"OpenOrders address {slot.spot_open_orders} at index {slot.index} not loaded."
                    )

                # Here's a comment from ckamm in https://github.com/blockworks-foundation/entropy-v3/pull/78/files
                # that describes some of the health calculations.
                #
                # // Two "worst-case" scenarios are considered:
                # // 1. All bids are executed at current price, producing a base amount of bids_base_net
                # //    when all quote_locked are converted to base.
                # // 2. All asks are executed at current price, producing a base amount of asks_base_net
                # //    because base_locked would be converted to quote.
                #
                # // Report the scenario that would have a worse outcome on health.
                # //
                # // Explanation: This function returns (base, quote) and the values later get used in
                # //     health += (if base > 0 { asset_weight } else { liab_weight }) * base + quote
                # // and here we return the scenario that will increase health the least.
                # //
                # // Correctness proof:
                # // - always bids_base_net >= asks_base_net
                # // - note that scenario 1 returns (a + b, c)
                # //         and scenario 2 returns (a,     c + b), and b >= 0, c >= 0
                # // - if a >= 0: scenario 1 will lead to less health as asset_weight <= 1.
                # // - if a < 0 and b <= -a: scenario 2 will lead to less health as liab_weight >= 1.
                # // - if a < 0 and b > -a:
                # //   The health contributions of both scenarios are identical if
                # //       asset_weight * (a + b) + c = liab_weight * a + c + b
                # //   <=> b = (asset_weight - liab_weight) / (1 - asset_weight) * a
                # //   <=> b = -2 a  since asset_weight + liab_weight = 2 by weight construction
                # //   So the worse scenario switches when a + b = -a.
                # // That means scenario 1 leads to less health whenever |a + b| > |a|.

                # base total if all bids were executed
                spot_bids_base_net = (
                    slot.net_value.value
                    + (spot_open_orders.quote_token_locked / price.value)
                    + spot_open_orders.base_token_total
                )

                # base total if all asks were executed
                spot_asks_base_net = (
                    slot.net_value.value + spot_open_orders.base_token_free
                )

                if spot_bids_base_net.copy_abs() > spot_asks_base_net.copy_abs():
                    spot_health_base = spot_bids_base_net
                    spot_health_quote = spot_open_orders.quote_token_free
                else:
                    spot_health_base = spot_asks_base_net
                    spot_health_quote = (
                        spot_open_orders.base_token_locked * price.value
                    ) + spot_open_orders.quote_token_total

            # From Daffy in Discord 2021-11-23: https://discord.com/channels/791995070613159966/857699200279773204/912705017767677982
            # --
            # There's a long_funding field on the PerpMarketCache which holds the current native USDC per
            # base position accrued. The long_settled_funding stores the last time funding was settled for
            # this particular user. So the funding owed is
            #   (PerpMarketCache.long_funding - PerpAccount.long_settled_funding) * PerpAccount.base_position
            # if base position greater than 0 (i.e. long)
            #
            # And we use short_funding if base_position < 0
            #
            # The long_funding field in PerpMarketCache changes across time according to the
            # update_funding() function. If orderbook is above index price, then long_funding and
            # short_funding both increase.
            #
            # Usually long_funding and short_funding will be the same unless there was a socialized loss
            # event. IF you have negative equity and insurance fund is empty, then half of the negative
            # equity goes to longs and half goes to shorts. The way that's done is by increasing
            # long_funding and decreasing short_funding by same amount.
            #
            # But unless there's a socialized loss, long_funding == short_funding
            # --
            perp_position: Decimal = Decimal(0)
            perp_notional_position: Decimal = Decimal(0)
            perp_health_base: Decimal = Decimal(0)
            perp_health_quote: Decimal = Decimal(0)
            unsettled_funding: Decimal = Decimal(0)
            perp_health_base_value: Decimal = Decimal(0)
            perp_asset: Decimal = Decimal(0)
            perp_liability: Decimal = Decimal(0)
            redeemable_pnl: Decimal = Decimal(0)
            perp_base_lot_size: Decimal = Decimal(0)
            perp_quote_lot_size: Decimal = Decimal(0)
            if (
                slot.perp_account is not None
                and not slot.perp_account.empty
                and market_cache is not None
            ):
                perp_market: typing.Optional[
                    GroupSlotPerpMarket
                ] = group.perp_markets_by_index[slot.index]
                if perp_market is None:
                    raise Exception(
                        f"Could not find perp market in Group at index {slot.index}."
                    )

                perp_base_lot_size = perp_market.base_lot_size
                perp_quote_lot_size = perp_market.quote_lot_size

                perp_position = (
                    slot.perp_account.lot_size_converter.base_size_lots_to_number(
                        slot.perp_account.base_position
                    )
                )
                perp_notional_position = perp_position * price.value
                cached_perp_market: typing.Optional[
                    PerpMarketCache
                ] = market_cache.perp_market
                if cached_perp_market is None:
                    raise Exception(
                        f"Could not find perp market in Cache at index {slot.index}."
                    )

                unsettled_funding = slot.perp_account.unsettled_funding(
                    cached_perp_market
                )
                bids_quantity = (
                    slot.perp_account.lot_size_converter.base_size_lots_to_number(
                        slot.perp_account.bids_quantity
                    )
                )
                asks_quantity = (
                    slot.perp_account.lot_size_converter.base_size_lots_to_number(
                        slot.perp_account.asks_quantity
                    )
                )
                taker_quote = (
                    slot.perp_account.lot_size_converter.quote_size_lots_to_number(
                        slot.perp_account.taker_quote
                    )
                )

                perp_bids_base_net: Decimal = perp_position + bids_quantity
                perp_asks_base_net: Decimal = perp_position - asks_quantity

                quote_pos = slot.perp_account.quote_position / (
                    10**self.shared_quote_token.decimals
                )
                if perp_bids_base_net.copy_abs() > perp_asks_base_net.copy_abs():
                    perp_health_base = perp_bids_base_net
                    perp_health_quote = (
                        (quote_pos + unsettled_funding)
                        + taker_quote
                        - (bids_quantity * price.value)
                    )
                else:
                    perp_health_base = perp_asks_base_net
                    perp_health_quote = (
                        (quote_pos + unsettled_funding)
                        + taker_quote
                        + (asks_quantity * price.value)
                    )
                perp_health_base_value = perp_health_base * price.value

                perp_asset = slot.perp_account.asset_value(
                    cached_perp_market, price.value
                )
                perp_liability = slot.perp_account.liability_value(
                    cached_perp_market, price.value
                )

                redeemable_pnl = slot.perp_account.pnl(cached_perp_market, price)

            group_slot: typing.Optional[GroupSlot] = None
            if market_cache is not None:
                group_slot = group.slot_by_instrument(slot.base_instrument)

            spot_init_asset_weight: Decimal = Decimal(0)
            spot_maint_asset_weight: Decimal = Decimal(0)
            spot_init_liab_weight: Decimal = Decimal(0)
            spot_maint_liab_weight: Decimal = Decimal(0)
            if group_slot is not None and group_slot.spot_market is not None:
                spot_init_asset_weight = group_slot.spot_market.init_asset_weight
                spot_maint_asset_weight = group_slot.spot_market.maint_asset_weight
                spot_init_liab_weight = group_slot.spot_market.init_liab_weight
                spot_maint_liab_weight = group_slot.spot_market.maint_liab_weight
            elif slot.base_instrument == self.shared_quote_token:
                spot_init_asset_weight = Decimal(1)
                spot_maint_asset_weight = Decimal(1)
                spot_init_liab_weight = Decimal(1)
                spot_maint_liab_weight = Decimal(1)

            perp_init_asset_weight: Decimal = Decimal(0)
            perp_maint_asset_weight: Decimal = Decimal(0)
            perp_init_liab_weight: Decimal = Decimal(0)
            perp_maint_liab_weight: Decimal = Decimal(0)
            if group_slot is not None and group_slot.perp_market is not None:
                perp_init_asset_weight = group_slot.perp_market.init_asset_weight
                perp_maint_asset_weight = group_slot.perp_market.maint_asset_weight
                perp_init_liab_weight = group_slot.perp_market.init_liab_weight
                perp_maint_liab_weight = group_slot.perp_market.maint_liab_weight
            elif slot.base_instrument == self.shared_quote_token:
                perp_init_asset_weight = Decimal(1)
                perp_maint_asset_weight = Decimal(1)
                perp_init_liab_weight = Decimal(1)
                perp_maint_liab_weight = Decimal(1)

            base_open_unsettled: Decimal = Decimal(0)
            base_open_locked: Decimal = Decimal(0)
            base_open_total: Decimal = Decimal(0)
            base_open_total_value: Decimal = Decimal(0)
            quote_open_unsettled: Decimal = Decimal(0)
            quote_open_locked: Decimal = Decimal(0)
            if spot_open_orders is not None:
                if (
                    slot.index < len(self.in_margin_basket)
                    and self.in_margin_basket[slot.index]
                ):
                    base_open_unsettled = spot_open_orders.base_token_free
                    base_open_locked = spot_open_orders.base_token_locked
                    base_open_total = spot_open_orders.base_token_total
                    base_open_total_value = base_open_total * price.value

                # Some calculations include quote unsettled whether it's in
                # the margin basket or not.
                quote_open_unsettled = (
                    spot_open_orders.quote_token_free
                    + spot_open_orders.referrer_rebate_accrued
                )
                quote_open_locked = spot_open_orders.quote_token_locked
            base_total: Decimal = (
                slot.deposit.value - slot.borrow.value + base_open_total
            )

            base_total_value: Decimal = base_total * price.value
            data = {
                "Name": slot.base_instrument.name,
                "Symbol": slot.base_instrument.symbol,
                "InMarginBasket": slot.index < len(self.in_margin_basket)
                and self.in_margin_basket[slot.index],
                "CurrentPrice": price.value,
                "Spot": base_total,
                "SpotValue": base_total_value,
                "SpotDeposit": slot.deposit.value,
                "SpotDepositValue": slot.deposit.value * price.value,
                "SpotBorrow": slot.borrow.value,
                "SpotBorrowValue": slot.borrow.value * price.value * Decimal(-1),
                "SpotOpen": base_open_total,
                "SpotOpenValue": base_open_total_value,
                "BaseUnsettled": base_open_unsettled,
                "BaseLocked": base_open_locked,
                "BaseLockedValue": base_open_locked * price.value,
                "QuoteUnsettled": quote_open_unsettled,
                "QuoteLocked": quote_open_locked,
                "PerpPositionSize": perp_position,
                "PerpNotionalSize": perp_notional_position,
                "SpotHealthBase": spot_health_base,
                "SpotHealthBaseValue": spot_health_base * price.value,
                "SpotHealthQuote": spot_health_quote,
                "PerpHealthBase": perp_health_base,
                "PerpHealthBaseValue": perp_health_base_value,
                "PerpHealthQuote": perp_health_quote,
                "PerpAsset": perp_asset,
                "PerpLiability": perp_liability,
                "RedeemablePnL": redeemable_pnl,
                "UnsettledFunding": unsettled_funding,
                "SpotInitAssetWeight": spot_init_asset_weight,
                "SpotMaintAssetWeight": spot_maint_asset_weight,
                "SpotInitLiabilityWeight": spot_init_liab_weight,
                "SpotMaintLiabilityWeight": spot_maint_liab_weight,
                "PerpInitAssetWeight": perp_init_asset_weight,
                "PerpMaintAssetWeight": perp_maint_asset_weight,
                "PerpInitLiabilityWeight": perp_init_liab_weight,
                "PerpMaintLiabilityWeight": perp_maint_liab_weight,
                "BaseDecimals": slot.base_instrument.decimals,
                "QuoteDecimals": group.shared_quote_token.decimals,
                "PerpBaseLotSize": perp_base_lot_size,
                "PerpQuoteLotSize": perp_quote_lot_size,
            }
            asset_data += [data]
        frame: pandas.DataFrame = pandas.DataFrame(asset_data)
        return frame

    def weighted_assets(
        self, frame: pandas.DataFrame, weighting_name: str = ""
    ) -> typing.Tuple[Decimal, Decimal]:
        non_quote = frame.loc[frame["Symbol"] != self.shared_quote_token.symbol]
        quote = frame.loc[
            frame["Symbol"] == self.shared_quote_token.symbol, "SpotValue"
        ].sum()
        quote += frame["PerpHealthQuote"].sum()

        # Sometimes there is QuoteUnsettled when the instrument is no longer in the margin
        # basket. Those values are excluded here to match the behaviour of the TypeScript
        # client so our answers match.
        quote += frame.loc[frame["InMarginBasket"], "QuoteUnsettled"].sum()

        assets = Decimal(0)
        liabilities = Decimal(0)
        if quote > 0:
            assets = quote
        else:
            liabilities = quote

        spot_borrow_health = (
            non_quote.loc[non_quote["SpotHealthBaseValue"] < 0, "SpotHealthBaseValue"]
            * non_quote[f"Spot{weighting_name}LiabilityWeight"]
        ).sum()

        perp_health_base_liability = (
            non_quote.loc[non_quote["PerpHealthBaseValue"] < 0, "PerpHealthBaseValue"]
            * non_quote[f"Perp{weighting_name}LiabilityWeight"]
        ).sum()

        liabilities += spot_borrow_health + perp_health_base_liability

        spot_deposit_health = (
            (non_quote.loc[non_quote["SpotHealthBaseValue"] > 0, "SpotHealthBaseValue"])
            * non_quote[f"Spot{weighting_name}AssetWeight"]
        ).sum()

        perp_health_base_asset = (
            non_quote.loc[non_quote["PerpHealthBaseValue"] > 0, "PerpHealthBaseValue"]
            * non_quote[f"Perp{weighting_name}AssetWeight"]
        ).sum()

        assets += spot_deposit_health + perp_health_base_asset

        return assets, liabilities

    def unweighted_assets(
        self, frame: pandas.DataFrame
    ) -> typing.Tuple[Decimal, Decimal]:
        non_quote = frame.loc[frame["Symbol"] != self.shared_quote_token.symbol]
        quote = frame.loc[
            frame["Symbol"] == self.shared_quote_token.symbol, "SpotValue"
        ].sum()

        assets = Decimal(0)
        liabilities = Decimal(0)
        if quote > 0:
            assets = quote
        else:
            liabilities = quote

        liabilities += (
            non_quote["SpotBorrowValue"].sum() + non_quote["PerpLiability"].sum()
        )

        assets += (
            non_quote["SpotDepositValue"].sum()
            + non_quote["BaseLockedValue"].sum()
            + non_quote["PerpAsset"].sum()
            + non_quote["QuoteUnsettled"].sum()
            + non_quote["QuoteLocked"].sum()
        )

        return assets, liabilities

    def init_health(self, frame: pandas.DataFrame) -> InstrumentValue:
        assets, liabilities = self.weighted_assets(frame, "Init")
        value: Decimal = assets + liabilities
        return InstrumentValue(self.shared_quote_token, value)

    def maint_health(self, frame: pandas.DataFrame) -> InstrumentValue:
        assets, liabilities = self.weighted_assets(frame, "Maint")
        value: Decimal = assets + liabilities
        return InstrumentValue(self.shared_quote_token, value)

    def init_health_ratio(self, frame: pandas.DataFrame) -> Decimal:
        assets, liabilities = self.weighted_assets(frame, "Init")
        if liabilities == 0:
            return Decimal(100)

        return ((assets / -liabilities) - 1) * 100

    def maint_health_ratio(self, frame: pandas.DataFrame) -> Decimal:
        assets, liabilities = self.weighted_assets(frame, "Maint")
        if liabilities == 0:
            return Decimal(100)

        return ((assets / -liabilities) - 1) * 100

    def total_value(self, frame: pandas.DataFrame) -> InstrumentValue:
        assets, liabilities = self.unweighted_assets(frame)

        value: Decimal = assets + liabilities
        return InstrumentValue(self.shared_quote_token, value)

    def is_liquidatable(self, frame: pandas.DataFrame) -> bool:
        if self.being_liquidated and self.init_health(frame) < 0:
            return True
        elif self.maint_health(frame) < 0:
            return True
        return False

    def leverage(self, frame: pandas.DataFrame) -> Decimal:
        assets, liabilities = self.unweighted_assets(frame)
        if assets <= 0:
            return Decimal(0)
        return -liabilities / (assets + liabilities)

    def redeemable_pnl(self, frame: pandas.DataFrame) -> InstrumentValue:
        value: Decimal = frame["RedeemablePnL"].sum()
        return InstrumentValue(self.shared_quote_token, value)

    def __str__(self) -> str:
        info = f"'{self.info}'" if self.info else "(un-named)"
        shared_quote: str = f"{self.shared_quote}".replace("\n", "\n        ")
        slot_count = len(self.base_slots)
        slots = "\n        ".join(
            [f"{item}".replace("\n", "\n        ") for item in self.base_slots]
        )

        symbols: typing.Sequence[str] = [
            slot.base_instrument.symbol for slot in self.base_slots
        ]
        in_margin_basket = ", ".join(symbols) or "None"
        return f"""« Account {info}, {self.version} [{self.address}]
    {self.meta_data}
    Owner: {self.owner}
    Delegated To: {self.delegate}
    Group: « Group '{self.group_name}' [{self.group_address}] »
    Advanced Orders Account: {self.advanced_orders}
    MSRM: {self.msrm_amount}
    Bankrupt? {self.is_bankrupt}
    Upgradable? {not self.not_upgradable}
    Being Liquidated? {self.being_liquidated}
    Shared Quote Token:
        {shared_quote}
    In Basket: {in_margin_basket}
    Basket [{slot_count} in basket]:
        {slots}
»"""

    def __repr__(self) -> str:
        return f"{self}"


@dataclass
class Valuation:
    account: Account
    open_orders: typing.Dict[str, OpenOrders]
    frame: pandas.DataFrame
    total_value: InstrumentValue

    def __init__(
        self,
        account: Account,
        open_orders: typing.Dict[str, OpenOrders],
        frame: pandas.DataFrame,
    ) -> None:
        self.account = account
        self.open_orders = open_orders
        self.frame = frame
        self.total_value = account.total_value(self.frame)

    @property
    def init_health(self) -> InstrumentValue:
        return self.account.init_health(self.frame)

    @property
    def maint_health(self) -> InstrumentValue:
        return self.account.maint_health(self.frame)

    @property
    def init_health_ratio(self) -> Decimal:
        return self.account.init_health_ratio(self.frame)

    @property
    def maint_health_ratio(self) -> Decimal:
        return self.account.maint_health_ratio(self.frame)

    @property
    def is_liquidatable(self) -> bool:
        return self.account.is_liquidatable(self.frame)

    @property
    def leverage(self) -> Decimal:
        return self.account.leverage(self.frame)

    @property
    def redeemable_pnl(self) -> InstrumentValue:
        return self.account.redeemable_pnl(self.frame)

    @staticmethod
    def create(
        account: Account,
        group: Group,
        open_orders_data: typing.Dict[str, AccountInfo],
        cache: Cache,
    ) -> "Valuation":
        oos: typing.Dict[str, OpenOrders] = {}
        for slot in account.slots:
            if slot.spot_open_orders is not None:
                oo = OpenOrders.parse(
                    open_orders_data[str(slot.spot_open_orders)],
                    Token.ensure(slot.base_instrument),
                    group.shared_quote_token,
                )
                oos[str(oo.address)] = oo
        frame = account.to_dataframe(group, oos, cache)
        return Valuation(account, oos, frame)

    @staticmethod
    def load(context: Context, address: PublicKey) -> "Valuation":
        group = Group.load(context)
        cache = Cache.load(context, group.cache)
        account = Account.load(context, address, group)
        open_orders = account.load_all_spot_open_orders(context)
        frame = account.to_dataframe(group, open_orders, cache)

        return Valuation(account, open_orders, frame)

    def __lt__(self, other: "Valuation") -> bool:
        return self.total_value.value < other.total_value.value

    def __le__(self, other: "Valuation") -> bool:
        return self.total_value.value <= other.total_value.value

    def __gt__(self, other: "Valuation") -> bool:
        return self.total_value.value > other.total_value.value

    def __ge__(self, other: "Valuation") -> bool:
        return self.total_value.value >= other.total_value.value

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Valuation):
            return self.total_value.value == other.total_value.value
        return False

    def __ne__(self, other: object) -> bool:
        if isinstance(other, Valuation):
            return self.total_value.value != other.total_value.value
        return True

    def __str__(self) -> str:
        return f"""« Valuation {str(self.account.address):<44} {self.total_value.value:>25,.8f} {self.total_value.token.symbol} »"""

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://org-docs.friktion.fi/entropy/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/Friktion-Labs)
#   [Email](mailto:hello@blockworks.foundation)

import jsons

from decimal import Decimal

from .tokens import Instrument


# # 🥭 LotSizeConverter class
#
class LotSizeConverter:
    def __init__(
        self,
        base: Instrument,
        base_lot_size: Decimal,
        quote: Instrument,
        quote_lot_size: Decimal,
    ) -> None:
        self.base: Instrument = base
        self.base_lot_size: Decimal = base_lot_size
        self.quote: Instrument = quote
        self.quote_lot_size: Decimal = quote_lot_size

    @property
    def lot_size(self) -> Decimal:
        return self.base_size_lots_to_number(Decimal(1))

    @property
    def tick_size(self) -> Decimal:
        return self.price_lots_to_number(Decimal(1))

    def adjust_to_base_decimals(self, value: Decimal) -> Decimal:
        adjusted = 10 ** (self.base.decimals - self.quote.decimals)
        return value * adjusted

    def adjust_to_quote_decimals(self, value: Decimal) -> Decimal:
        adjusted = 10 ** (self.quote.decimals - self.base.decimals)
        return value * adjusted

    def price_lots_to_number(self, price_lots: Decimal) -> Decimal:
        lots_to_native = self.quote_lot_size / self.base_lot_size
        return self.adjust_to_base_decimals(price_lots * lots_to_native)

    def price_number_to_lots(self, price: Decimal) -> int:
        base_factor: Decimal = 10**self.base.decimals
        quote_factor: Decimal = 10**self.quote.decimals
        return round(
            (price * quote_factor * self.base_lot_size)
            / (base_factor * self.quote_lot_size)
        )

    def base_size_lots_to_number(self, size_lots: Decimal) -> Decimal:
        size: int = round(size_lots)
        base_factor: Decimal = 10**self.base.decimals
        return Decimal(size * self.base_lot_size) / base_factor

    def base_size_number_to_lots(self, size: Decimal) -> int:
        base_factor: Decimal = 10**self.base.decimals
        return int(round(size * base_factor) / self.base_lot_size)

    def quote_size_lots_to_number(self, size_lots: Decimal) -> Decimal:
        size: int = round(size_lots)
        quote_factor: Decimal = 10**self.quote.decimals
        return Decimal(size * self.quote_lot_size) / quote_factor

    def quote_lots_to_number(self, size_lots: Decimal) -> Decimal:
        quote_factor: Decimal = 10**self.quote.decimals
        return Decimal(size_lots * self.quote_lot_size) / quote_factor

    def quote_size_number_to_lots(self, size: Decimal) -> int:
        quote_factor: Decimal = 10**self.quote.decimals
        return int(round(size * quote_factor) / self.quote_lot_size)

    def round_base(self, quantity: Decimal) -> Decimal:
        return round(quantity / self.lot_size) * self.lot_size

    def round_quote(self, price: Decimal) -> Decimal:
        return round(price / self.tick_size) * self.tick_size

    def __str__(self) -> str:
        return f"« LotSizeConverter {self.base.symbol}/{self.quote.symbol} [base lot size: {self.base_lot_size} ({self.base.decimals} decimals), quote lot size: {self.quote_lot_size} ({self.quote.decimals} decimals)] »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 NullLotSizeConverter class
#
class NullLotSizeConverter(LotSizeConverter):
    def __init__(self) -> None:
        super().__init__(
            Instrument("NULLBASE", "Null Base", Decimal(0)),
            Decimal(1),
            Instrument("NULLQUOTE", "Null Quote", Decimal(0)),
            Decimal(1),
        )

    def price_lots_to_number(self, price_lots: Decimal) -> Decimal:
        return price_lots

    def price_number_to_lots(self, price: Decimal) -> int:
        return round(price)

    def base_size_lots_to_number(self, size_lots: Decimal) -> Decimal:
        return size_lots

    def base_size_number_to_lots(self, size: Decimal) -> int:
        return round(size)

    def quote_size_lots_to_number(self, size_lots: Decimal) -> Decimal:
        return size_lots

    def quote_size_number_to_lots(self, size: Decimal) -> int:
        return round(size)

    def __str__(self) -> str:
        return "« NullLotSizeConverter »"


# # 🥭 RaisingLotSizeConverter class
#
class RaisingLotSizeConverter(LotSizeConverter):
    def __init__(self) -> None:
        super().__init__(
            Instrument("RAISINGBASE", "Raising Base", Decimal(0)),
            Decimal(-1),
            Instrument("RAISINGQUOTE", "Raising Quote", Decimal(0)),
            Decimal(-1),
        )

    def price_lots_to_number(self, price_lots: Decimal) -> Decimal:
        raise NotImplementedError(
            "RaisingLotSizeConverter.price_lots_to_number() is not implemented. RaisingLotSizeConverter is a stub used where no LotSizeConverter members should be called."
        )

    def price_number_to_lots(self, price: Decimal) -> int:
        raise NotImplementedError(
            "RaisingLotSizeConverter.price_number_to_lots() is not implemented. RaisingLotSizeConverter is a stub used where no LotSizeConverter members should be called."
        )

    def base_size_lots_to_number(self, size_lots: Decimal) -> Decimal:
        raise NotImplementedError(
            "RaisingLotSizeConverter.base_size_lots_to_number() is not implemented. RaisingLotSizeConverter is a stub used where no LotSizeConverter members should be called."
        )

    def base_size_number_to_lots(self, size: Decimal) -> int:
        raise NotImplementedError(
            "RaisingLotSizeConverter.base_size_number_to_lots() is not implemented. RaisingLotSizeConverter is a stub used where no LotSizeConverter members should be called."
        )

    def quote_size_lots_to_number(self, size_lots: Decimal) -> Decimal:
        raise NotImplementedError(
            "RaisingLotSizeConverter.quote_size_lots_to_number() is not implemented. RaisingLotSizeConverter is a stub used where no LotSizeConverter members should be called."
        )

    def quote_size_number_to_lots(self, size: Decimal) -> int:
        raise NotImplementedError(
            "RaisingLotSizeConverter.quote_size_number_to_lots() is not implemented. RaisingLotSizeConverter is a stub used where no LotSizeConverter members should be called."
        )

    def __str__(self) -> str:
        return "« RaisingLotSizeConverter »"


jsons.set_serializer(lambda _, **__: "{}", RaisingLotSizeConverter)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import logging
import typing

from solana.publickey import PublicKey

from .instrumentvalue import InstrumentValue


# # 🥭 OwnedInstrumentValue class
#
# Ties an owner and `InstrumentValue` together. This is useful in the `TransactionScout`, where
# token mints and values are given separate from the owner `PublicKey` - we can package them
# together in this `OwnedInstrumentValue` class.


class OwnedInstrumentValue:
    def __init__(self, owner: PublicKey, token_value: InstrumentValue) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.owner = owner
        self.token_value = token_value

    @staticmethod
    def find_by_owner(
        values: typing.Sequence["OwnedInstrumentValue"], owner: PublicKey
    ) -> "OwnedInstrumentValue":
        found = [value for value in values if value.owner == owner]
        if len(found) == 0:
            raise Exception(f"Owner '{owner}' not found in: {values}")

        if len(found) > 1:
            raise Exception(f"Owner '{owner}' matched multiple tokens in: {values}")

        return found[0]

    @staticmethod
    def changes(
        before: typing.Sequence["OwnedInstrumentValue"],
        after: typing.Sequence["OwnedInstrumentValue"],
    ) -> typing.Sequence["OwnedInstrumentValue"]:
        changes: typing.List[OwnedInstrumentValue] = []
        for before_value in before:
            after_value = OwnedInstrumentValue.find_by_owner(after, before_value.owner)
            token_value = InstrumentValue(
                before_value.token_value.token,
                after_value.token_value.value - before_value.token_value.value,
            )
            result = OwnedInstrumentValue(before_value.owner, token_value)
            changes += [result]

        return changes

    def __str__(self) -> str:
        return f"« OwnedInstrumentValue [{self.owner}]: {self.token_value} »"

    def __repr__(self) -> str:
        return f"{self}"
